package dk.brics.grammar;

/**
 * Grammar production entity.
 */
public abstract class Entity {
	
	/** Label of entity, null if ignorable. */
	private String label;

	/** Example string, null if absent. */
	private String example;
	
	/**
	 * Constructs a new entity.
	 * @param label entity label, null if ignorable
	 * @param example example string, null if absent
	 * @throws GrammarException if example string does not match the entity
	 */
	protected Entity(String label, String example) throws GrammarException {
		this.label = label;
		this.example = example;
	}

	/**
	 * Returns true if this entity is labeled.
	 * @return true if this entity is labeled
	 * @see #isExplicitlyLabeled()
	 */
	public boolean isLabeled() {
		return label != null;
	}
	
	/**
	 * Returns the example string
	 * @return example string, null if absent
	 */
	public String getExample() {
		return example;
	}

	/**
	 * Returns the label of this entity. 
	 * @return label, null if absent
	 */
	public String getLabel() {
		return label;
	}
	/**
	 * Sets the label.
	 * @param label new label
	 */
	public void setLabel(String label) {
		this.label = label;
	}
	
	/**
	 * Checks whether the label is explicit or auto-generated.
	 * @return true if explicit
	 * @see #isLabeled()
	 */
	public boolean isExplicitlyLabeled() {
		return label != null && !label.startsWith("#");
	}
	
	/**
	 * Visits this entity with the given visitor.
	 * @param v entity visitor
	 */
	abstract public <T> T visitBy(EntityVisitor<T> v);
}
