package dk.brics.grammar;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Map;
import java.util.TreeSet;

import static dk.brics.misc.Chars.escape;

/**
 * Context-free grammar.
 */
public class Grammar {
	
	/** Start nonterminal. */
	private String start;

	/** Productions. */
	private Collection<Production> productions;
	
	/** Map from nonterminal name to productions. */
	private Map<String, Collection<Production>> productionmap;
	
	/** Map from production IDs to production objects. */
	private Map<ProductionID, Production> idmap;
	
	/** Nonterminals that have productions */
	private Collection<String> nonterminals;
	
	/** Set if this grammar results from unfolding. */
	private boolean unfolded;

	/**
	 * Constructs a new grammar.
	 * Also sets default names for productions and (if no entities are labeled) for entities.
	 * @param start start nonterminal
	 * @param productions collection of productions
	 */
	public Grammar(String start, Collection<Production> productions) {
		this.start = start;
		this.productions = productions;
		productionmap = new HashMap<String, Collection<Production>>();
		idmap = new HashMap<ProductionID,Production>();
		nonterminals = new TreeSet<String>();
		boolean add_entitylabels = true;
		for (Production p : productions) {
			idmap.put(p.getID(), p);
			Collection<Production> c = productionmap.get(p.getNonterminal());
			if (c == null) {
				c = new ArrayList<Production>();
				productionmap.put(p.getNonterminal(), c);
				nonterminals.add(p.getNonterminal());
			}
			c.add(p);
			ProductionID id = p.getID();
			if (id.getLabel() == null) // set dummy production labels
				id.setLabel("#" + c.size());
			for (Entity e : p.getEntities())
				if (e.getLabel() != null)
					add_entitylabels = false;
		}
		if (add_entitylabels) // set dummy entity labels
			for (Production p : productions)  {
				int i = 1;
				for (Entity e : p.getEntities())
					if (e instanceof NonterminalEntity || e instanceof RegexpTerminalEntity)
						e.setLabel("#" + i++);
			}
	}

	/**
	 * Constructs a copy of the given grammar.
	 * Production objects are cloned.
	 */
	public Grammar(Grammar g) {
		this(g.getStart(), g.cloneProductions());
	}
	
	private Collection<Production> cloneProductions() {
		Collection<Production> newprods = new ArrayList<Production>();
		for (Production p : productions)
			newprods.add(new Production(p.getNonterminal(), p.getEntities(), p.isUnordered(), p.getID(), p.getPriority()));
		return newprods;
	}
	
	/**
	 * Adds new productions to this grammar.
	 * @param newprods new productions
	 */
	public void addProductions(Collection<Production> newprods) {
		productions.addAll(newprods);
		for (Production p : productions) {
			idmap.put(p.getID(), p);
			Collection<Production> c = productionmap.get(p.getNonterminal());
			if (c == null) {
				c = new ArrayList<Production>();
				productionmap.put(p.getNonterminal(), c);
				nonterminals.add(p.getNonterminal());
			}
			c.add(p);
		}
	}
	
	/**
	 * Gets production with the given ID.
	 * @param id production ID
	 * @return production
	 */
	public Production getProduction(ProductionID id) {
		return idmap.get(id);
	}
	
	/**
	 * Gets productions with the given left-hand-side nonterminal.
	 * @param nonterminal nonterminal
	 * @return productions
	 */
	public Collection<Production> getProductions(String nonterminal) {
		return productionmap.get(nonterminal);
	}

	/**
	 * Returns the productions.
	 * @return collection of productions.
	 */
	public Collection<Production> getProductions() {
		return productions;
	}

	/**
	 * Returns the start nonterminal.
	 * @return nonterminal
	 */
	public String getStart() {
		return start;
	}
	
	/**
	 * Sets the start nonterminal.
	 */
	public void setStart(String start) {
		this.start = start;
	}
	
	/**
	 * Returns the nonterminals that have productions.
	 * @return collection of nonterminals
	 */
	public Collection<String> getNonterminals() {
		return nonterminals;
	}

	/**
	 * Returns string representation of the productions in this grammar.
	 * The start nonterminal comes first. Regexp definitions are not included.
	 */
	@Override
	public String toString() {
		StringBuilder b = new StringBuilder();
		Production[] ordered = productions.toArray(new Production[0]);
		Arrays.sort(ordered, new Comparator<Production>() {
			public int compare(Production p1, Production p2) { // not consistent with equals
				int c = p1.getNonterminal().compareTo(p2.getNonterminal());
				if (c == 0)
					c = p2.getPriority() - p1.getPriority();
				else {
					if (p1.getNonterminal().equals(start))
						c = -1;
					else if (p2.getNonterminal().equals(start))
						c = 1;
				}
				if (c == 0 && p1.getID().hasExplicitLabel() && p2.getID().hasExplicitLabel())
					c = p1.getID().getLabel().compareTo(p2.getID().getLabel());
				return c;
			}
		});
		Production prev = null;
		for (Production p : ordered) {
			boolean hasname = false;
			if (prev == null || !prev.getNonterminal().equals(p.getNonterminal())) {
				b.append(p.getNonterminal());
				hasname = true;
			} else
				b.append(' ');
			if (p.getID().hasExplicitLabel())
				b.append('[').append(p.getID().getLabel()).append("]");
			if (hasname || p.getID().hasExplicitLabel())
				b.append(" ");
			if (prev != null && prev.getNonterminal().equals(p.getNonterminal()) && prev.getPriority() > p.getPriority())
				b.append(">");
			if (hasname)
				b.append(":");
			else
				b.append("|");
			if (p.unordered)
				b.append('&');
			for (Entity e : p.getEntities()) {
				b.append(' ').append(e);
				boolean haslabel = e.isExplicitlyLabeled();
				String example = e.getExample();
				if (haslabel || example != null)
					b.append('[');
				if (haslabel)
					b.append( e.getLabel());
				if (haslabel && example != null)
					b.append(' ');
				if (example != null)
					b.append('"').append(escape(example)).append('"');
				if (haslabel || example != null)
					b.append(']');
			}
			b.append('\n');
			prev = p;
		}
		return b.toString();
	}
	
	/**
	 * Sets the unfolded status of this grammar (default: false).
	 * @param unfolded new unfolded status
	 */
	public void setUnfolded(boolean unfolded) {
		this.unfolded = unfolded;
	}
	
	/**
	 * Returns true if this grammar results from unfolding.
	 * (Such grammars have expanded alphabet.)
	 * @return true if unfolded
	 */
	public boolean isUnfolded() {
		return unfolded;
	}
	
	/**
	 * add production to productions
	 * @param p
	 * @autor Jn Ruhalovsk
	 */
	public void addProduction (Production p, boolean isStart) {
		if (!isStart) {
			productions.add(p);
		}
		else {
			System.out.println("SIZE OF PRODUCTIOPNS: (before)" + productions.size());
			Collection<Production> temp_productions = new ArrayList<Production>();
			nonterminals.add(p.getNonterminal());
			temp_productions.add(p);
			temp_productions.addAll(productions);
			productions.clear();
			nonterminals.clear();//??
			idmap.clear();//??
			productionmap.clear();
			addProductions(temp_productions);
			//productions.addAll(temp_productions);
			System.out.println("SIXE OF PRODUCTIOPNS: (after)" + productions.size());
			for (Production pp : productions) {
				System.out.println(pp.getNonterminal() + " rules: "  + pp.getEntities().toString());
			}
		}
	}
}
