package dk.brics.grammar;

import java.util.Collection;
import java.util.List;

/**
 * A grammar production.
 */
public class Production {
	
	/** Left-hand-side nonterminal. */
	private String nonterminal;

	/** Right-hand-side entities. */
	private List<Entity> entities;
	
	/** Unordered/ordered flag. */
	boolean unordered;
	
	/** ID of this production. */
	private ProductionID id;
	
	/** Priority comparared to other productions of the same nonterminal. */
	private int priority;
	
	/** If set, this production results from grammar unfolding. */
	private boolean unfolded;

	/**
	 * Constructs a new grammar production.
	 * @param nonterminal nonterminal
	 * @param entities entity list
	 * @param unordered true if 'unordered' production, false if 'ordered' production 
	 * @param id production ID
	 * @param priority priority
	 */
	public Production(String nonterminal, List<Entity> entities, boolean unordered, ProductionID id, int priority) {
		this.nonterminal = nonterminal;
		this.entities = entities;
		this.unordered = unordered;
		this.id = id;
		this.priority = priority;
	}
	
	/**
	 * Prints this production.
	 * @param remaining remaining entities
	 * @return string representation
	 */
	public String print(Collection<Entity> remaining) {
		StringBuilder b = new StringBuilder();
		b.append(nonterminal).append("->");
		boolean first = true;
		for (Entity e : entities) {
			if (first)
				first = false;
			else
				b.append(",");
			if (remaining.contains(e))
				b.append(e);
			else
				b.append("(").append(e).append(")");
		}
		return b.toString();
	}
	
	/**
	 * Returns the left-hand-side nonterminal.
	 * @return nonterminal
	 */
	public String getNonterminal() {
		return nonterminal;
	}

	/**
	 * Returns true if this is a unordered production.
	 * @return true if this is a unordered production
	 */
	public boolean isUnordered() {
		return unordered;
	}

	/**
	 * Returns the entities.
	 * @return entities
	 */
	public List<Entity> getEntities() {
		return entities;
	}
	
	/**
	 * Sets the entities.
	 * @param entities new entities
	 */
	public void setEntities(List<Entity> entities) {
		this.entities = entities;
	}

	/**
	 * Returns the production ID.
	 * @return production ID
	 */
	public ProductionID getID() {
		return id;
	}
	
	/** 
	 * Returns the priority of this production.
	 * @return priority
	 */
	public int getPriority() {
		return priority;
	}
	
	/**
	 * Marks this production as one that results from grammr unfolding.
	 */
	public void setUnfolded() {
		unfolded = true;
	}

	/**
	 * Checks whether this production results from grammar unfolding.
	 * (Such productions are not checked for ambiguity.)
	 */
	public boolean isUnfolded() {
		return unfolded;
	}
	
	/**
	 * Compare productions
	 * return true, if are eqal
	 * return false, if aren't equal
	 * @param p compared production
	 * 
	 * @author Jn Ruhalovsk
	 */
	public boolean equalsTo(Production p) {
		if (!nonterminal.equals(p.getNonterminal()))
			return false;
		if (entities.size() != p.getEntities().size()) 
			return false;
		
		for (int i=0; i<entities.size(); i++) {
			if (!entities.get(i).equals(p.getEntities().get(i))) {
				return false;
			}
		}
		return true;
	}
	
	/**
	 * @author Jn Ruhalovsk
	 */
	@Override
	public boolean equals(Object obj) {
		if (!(obj instanceof Production)) {
			return false;
		}
		if (!nonterminal.equals(((Production)obj).getNonterminal()))
			return false;
		if (entities.size() != ((Production)obj).getEntities().size()) 
			return false;
		
		for (int i=0; i<entities.size(); i++) {
			if (!entities.get(i).toString().equals(((Production)obj).getEntities().get(i).toString())) {
				//System.out.println(entities.get(i) + " ############### " + ((Production)obj).getEntities().get(i));
				return false;
			}
		}
		return true;
	}
	
	/**
	 * @author Jn Ruhalovsk
	 */
	@Override
	public int hashCode() {
		StringBuilder res = new StringBuilder();
		res.append(nonterminal);
		res.append(entities.toString());
		return res.toString().hashCode();
	}
}
