package dk.brics.grammar.ambiguity;

import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.List;

import dk.brics.grammar.EOFTerminalEntity;
import dk.brics.grammar.Entity;
import dk.brics.grammar.EntityVisitor;
import dk.brics.grammar.Grammar;
import dk.brics.grammar.NonterminalEntity;
import dk.brics.grammar.Production;
import dk.brics.grammar.RegexpTerminalEntity;
import dk.brics.grammar.StringTerminalEntity;
import dk.brics.grammar.parser.Parser;
import dk.brics.misc.Properties;

import static dk.brics.misc.Chars.escape;

/**
 * Grammar ambiguity checker.
 * <p>
 * Based on "<a href="http://www.daimi.au.dk/~amoeller/papers/ambiguity/">Analyzing 
 * Ambiguity of Context-Free Grammars</a>", by Claus Brabrand, Robert Giegerich, and 
 * Anders Mller.
 */
public class AmbiguityAnalyzer {
	
	/**
	 * @author Jn Ruhalovsk
	 */
	private ArrayList<ProductionPair<Production, Production>> testedProductions;
	
	/** 
	 * Grammar for the current analysis. 
	 */
	private Grammar g;
	
	
	/** 
	 * Print stream for messages. 
	 */
	private PrintWriter out;
	
	/**
	 * If set to true, verbose output is generated.
	 */
	private boolean verbose;
	
	/** 
	 * If set to true, debug info is generated. 
	 * (Enable dynamically with dk.brics.grammar.ambiguity.debug)
	 */
	private boolean debug;
	
	/** 
	 * If set to true, potentially ambiguous example strings are not checked using the parser. 
	 * Default: not set.
	 * (Enable dynamically with dk.brics.grammar.ambiguity.noparsecheck)
	 */
	private boolean noparsecheck;
	
	/**
	 * If set to true, unlabeled nonterminal/regexp entities are treated as ignorable
	 * in horizontal checks.
	 * Default: autodetect (set if the grammar contains at least one labeled entity).
	 * (Enable dynamically with dk.brics.grammar.ambiguity.ignorables)
	 */
	private boolean ignorable_entities;
	
	/**
	 * If set to true, pairs of unlabeled productions are skipped in vertical checks.
	 * Default: autodetect (set if the grammat contains at least one labeled production).
	 * (Enable dynamically with dk.brics.grammar.ambiguity.ignorables)
	 */
	private boolean ignorable_productions;
	
	/** Approximation strategies. */
	private List<ApproximationStrategy> approx;
	
	/**
	 * Parser, used for optional/certain checks.
	 */
	private Parser parser;
	
	/** Maximal memory usage during last analysis. */
	private long max_memory;
	
	/** Number of potential horizontal ambiguities in current analysis. */
	private int potential_horizontal_ambiguities;
	
	/** Number of potential vertical ambiguities in current analysis. */
	private int potential_vertical_ambiguities;
	
	/** Number of certain horizontal ambiguities in current analysis. */
	private int certain_horizontal_ambiguities;
	
	/** Number of certain vertical ambiguities in current analysis. */
	private int certain_vertical_ambiguities;
	
	/** Number of <code>OutOfMemoryError</code>s caught. */
	private int outofmemory;
	
	/**
	 * Constructs a new grammar analyzer with default approximation strategies.
	 * Default approximation strategies: 
	 * <code>terminal</code>, 
	 * <code>regular2</code>, 
	 * <code>regular</code>.
	 * This default behavior can be changed by setting the system property
	 * <code>dk.brics.grammar.ambiguity.strategies</code> to the list of classes implementing the strategies.
	 * @param out print writer for messages
	 * @param verbose if enabled, verbose output is generated
	 * @throws ClassNotFoundException if an approximation strategy class was not found
	 * @throws IllegalAccessException if an approximation strategy class or its nullary constructor is not accessible
	 * @throws InstantiationException if an approximation strategy class cannot be instantiated
	 */
	public AmbiguityAnalyzer(PrintWriter out, boolean verbose) throws InstantiationException, IllegalAccessException, ClassNotFoundException {
		this.verbose = verbose;
		List<ApproximationStrategy> a = new ArrayList<ApproximationStrategy>();
		String[] strategies = Properties.getStrings("dk.brics.grammar.ambiguity.strategies");
		if (strategies != null) {
			ClassLoader loader = this.getClass().getClassLoader(); 
			for (String s : strategies)
				a.add((ApproximationStrategy)loader.loadClass(s).newInstance());
		} else {
			ApproximationStrategy[] default_strategies = { 
				new TerminalApproximation(),
				new RegularApproximation2(),
				new RegularApproximation(),
				new RegularApproximationCycleBreaking(),
			}; // keep the list in the javadoc updated...
			for (ApproximationStrategy as : default_strategies)
				a.add(as);
		}
		init(a, out);
	}

	/**
	 * Constructs a new grammar analyzer with the given list of approximation strategies.
	 * @param approx approximation strategies
	 * @param out print writer for messages
	 */
	public AmbiguityAnalyzer(List<ApproximationStrategy> approx, PrintWriter out) { 
		init(approx, out);
	}
	
	private void init(List<ApproximationStrategy> approx, PrintWriter out) {
		this.approx = approx;
		this.out = out;
		debug = Properties.get("dk.brics.grammar.ambiguity.debug");
		noparsecheck = Properties.get("dk.brics.grammar.ambiguity.noparsecheck");
		if (debug) {
			if (noparsecheck)
				out.println("parse check disabled");
			out.print("approximation strategies:");
			for (ApproximationStrategy a : approx)
				out.print(" " + a.getName());
			out.println();
		}
	}

	/**
	 * Analyzes the given grammar.
	 * @param g grammar
	 * @return true if the grammar is definitely unambiguous
	 */
	public boolean analyze(Grammar g) {
		this.g = g;
		System.out.println("ANALYZOVANA GRAMATIKA: " + g.toString());
		potential_horizontal_ambiguities = potential_vertical_ambiguities = 
			certain_horizontal_ambiguities = certain_vertical_ambiguities = 0;
		outofmemory = 0;
		parser = new Parser(g, out);
		if (debug) 
			out.print("grammar:\n" + g.toString());
		if (Properties.get("dk.brics.grammar.ambiguity.ignorables"))
			ignorable_entities = ignorable_productions = true;
		else {
			ignorable_entities = ignorable_productions = false;
			for (Production p : g.getProductions()) {
				if (p.getID().hasExplicitLabel())
					ignorable_productions = true;
				if (!ignorable_entities)
					for (Entity e : p.getEntities())
						if (e.isExplicitlyLabeled()) {
							ignorable_entities = true;
							break;
						}
			}
		}
		if (debug)
			out.println("ignorable entities " + (ignorable_entities ? "enabled" : "disabled") +
					", ignorable productions " + (ignorable_productions ? "enabled" : "disabled"));
		for (ApproximationStrategy a : approx) {
			if (debug)
				out.println("initializing: " + a.getName());
			a.init(g, out, debug);
		}
		if (debug && g.isUnfolded())
			out.println("grammar has been unfolded");
		if (debug || verbose)
			out.println("checking vertical ambiguities...");
		analyzeVertical();
		if (debug || verbose)
			out.println("checking horizontal ambiguities...");
		analyzeHorizontal();
		
//		/*
//		 * exhaustive search
//		 * added by Jn Ruhalovsk
//		 */
//		if (certain_vertical_ambiguities + potential_vertical_ambiguities 
//				+ certain_horizontal_ambiguities + potential_horizontal_ambiguities + outofmemory == 0) {
//			return certain_vertical_ambiguities + potential_vertical_ambiguities 
//					+ certain_horizontal_ambiguities + potential_horizontal_ambiguities + outofmemory == 0;
//		}
//		else {//RUN Exhaustive search
//			ExhaustiveSearcher es = new ExhaustiveSearcher(g, testedProductions);
//			es.search(5);
//		}
		
		return certain_vertical_ambiguities + potential_vertical_ambiguities 
			+ certain_horizontal_ambiguities + potential_horizontal_ambiguities + outofmemory == 0;
		
		
	}
	
	/**
	 * Prints statistics for the approximation strategies from the last analysis.
	 * @param out print writer for output
	 */
	public void printStatistics(PrintWriter out) {
		out.println("ambiguities detected: " 
				+ certain_vertical_ambiguities + " certain + " + potential_vertical_ambiguities + " potential vertical; " 
				+ certain_horizontal_ambiguities + " certain + " + potential_horizontal_ambiguities + " potential horizontal");
		if (outofmemory > 0)
			out.println("ambiguity checks with resources exhausted: " + outofmemory);
		for (ApproximationStrategy a : approx)
			a.printStatistics(out);
	}

	// FIXME: handle unordered productions (see also ProductionLengthNormalizer) 
	// TODO: handle equality entities
	// TODO: optionally set timeout on each horiz/vert check?
	
	private void analyzeHorizontal() {
		for (Production p :g.getProductions()) 
			if (!p.isUnfolded()) {
				int length = p.getEntities().size();
				for (int index = 1; index < length; index++)
					try {
						if (ignorable_entities && !p.getEntities().get(index - 1).isLabeled() && !p.getEntities().get(index).isLabeled())
							continue;
						if (debug || verbose)
							out.println(" horizontal check: " + p.getNonterminal() + "[" + p.getID().getLabel() + "] at index " + index);
						HorizontalOverlapString overlap = ApproximationStrategy.HORIZONTAL_NOT_APPLICABLE;
						boolean certain = false;
						boolean first = true;
						boolean checked = false;
						for (ApproximationStrategy a : approx) {
							HorizontalOverlapString a_overlap = a.horizontalCheck(p, index);
							if (a_overlap == ApproximationStrategy.HORIZONTAL_NOT_APPLICABLE)
								continue;
							if (debug)
								out.println("  result from " + a.getName() + ": " + (a_overlap == null ? "no overlap" : "\"" + escape(decode(a_overlap.getString())) + "\""));
							if (a_overlap == null) {
								overlap = null;
								break;
							}
							boolean a_certain = false;
							boolean a_checked = false;
							if (!noparsecheck) {
								a_certain = parseCheckHorizontal(p, index, a_overlap);
								a_checked = true;
							}
							if (first || a_certain) {
								overlap = a_overlap;
								certain = a_certain;
								checked = a_checked;
								first = false;
							}
						}
						if (overlap != null) {
							if (!noparsecheck && !checked)
								certain = parseCheckHorizontal(p, index, overlap);
							out.println("*** " + (certain ? "" : "potential ") + "horizontal ambiguity: " + describeHorizontal(p, index));
							if (certain) {
								out.println("    ambiguous string: \"" + escape(decode(overlap.getString())) + "\"");
								out.println("    matched as \"" + escape(decode(overlap.getX())) + "\" <--> \"" + 
										escape(decode(overlap.getAY())) + "\" or \"" + escape(decode(overlap.getXA())) + 
										"\" <--> \"" + escape(decode(overlap.getY())) + "\"");
								certain_horizontal_ambiguities++;
							} else
								potential_horizontal_ambiguities++;
							/*
							 * addedd by Jn Ruhalovsk
							 */
//							ProductionPair<Production, Production> pp = new ProductionPair<Production, Production>(p, null);
//							testedProductions.add(pp);
						}
						updateMaxMemory();
					} catch (OutOfMemoryError e) {
						updateMaxMemory();
						if (debug)
							out.println("  out of memory");
						out.println("*** resources exhausted, aborting horizontal ambiguity check: " + describeHorizontal(p, index));
						outofmemory++;
					}
				for (ApproximationStrategy a : approx)
					a.horizontalDone();
			}
	}
	
	private String describeHorizontal(Production p, int index) {
		return  p.getNonterminal() + "[" + p.getID().getLabel() + "]:" + 
			printEntities(p, 0, index) + " <-->" + printEntities(p, index, p.getEntities().size());
	}

	private void analyzeVertical() {
		for (String nt : g.getNonterminals()) {
			List<Production> prods = new ArrayList<Production>(g.getProductions(nt));
			for (int i1 = 0; i1+1 < prods.size(); i1++) {
				Production p1 = prods.get(i1);
				if (p1.isUnfolded())
					continue;
				for (int i2 = i1+1; i2 < prods.size(); i2++) { 
					Production p2 = prods.get(i2);
					System.out.println("**********ANLAYZE VERTICAL**********");
					System.out.println("P1: " + p1.print(p1.getEntities()) + " P2: " + p2.print(p2.getEntities()));
					if (p2.isUnfolded())
						continue;
					if (p1.getPriority() != p2.getPriority())
						continue;
					if (ignorable_productions && !p1.getID().hasExplicitLabel() && !p2.getID().hasExplicitLabel())
						continue;
					try {
						if (debug || verbose)
							out.println(" vertical check: " + p1.getNonterminal() + "[" + p1.getID().getLabel() + "] vs. " 
									+ p2.getNonterminal() + "[" + p2.getID().getLabel() + "]");
						VerticalOverlapString overlap = ApproximationStrategy.VERTICAL_NOT_APPLICABLE;
						boolean certain = false;
						boolean first = true;
						boolean checked = false;
						for (ApproximationStrategy a : approx) {
							VerticalOverlapString a_overlap = a.verticalCheck(p1, p2);
							if (a_overlap == ApproximationStrategy.VERTICAL_NOT_APPLICABLE)
								continue;
							if (debug)
								out.println("  result from " + a.getName() + ": " + (a_overlap == null ? "no overlap" : "\"" + escape(decode(a_overlap.getString())) + "\""));
							if (a_overlap == null) {
								overlap = null;
								break;
							}
							boolean a_certain = false;
							boolean a_checked = false;
							if (!noparsecheck) {
								a_certain = parseCheckVertical(p1, p2, a_overlap);
								a_checked = true;
							}
							if (first || a_certain) {
								overlap = a_overlap;
								certain = a_certain;
								checked = a_checked;
								first = false;
							}
						}
						if (overlap != null) {
							if (!noparsecheck && !checked)
								certain = parseCheckVertical(p1, p2, overlap);
							out.println("*** " + (certain ? "" : "potential ") + "vertical ambiguity: " + describeVertical(nt, p1, p2));
							//System.out.println("RULES: " + p1.print(p1.getEntities()) + " RULES2: " + p2.print(p2.getEntities()));
							if (certain) {
								out.println("    ambiguous string: \"" + escape(decode(overlap.getString())) + "\"");
								certain_vertical_ambiguities++;
							} else
								potential_vertical_ambiguities++;
						}
						updateMaxMemory();
					} catch (OutOfMemoryError e) {
						updateMaxMemory();
						if (debug)
							out.println("  out of memory");
						out.println("*** resources exhausted, aborting vertical ambiguity check: " + describeVertical(nt, p1, p2));
						outofmemory++;
					}
				}
			}
			for (ApproximationStrategy a : approx)
				a.verticalDone();
		}
	}
	
	private String describeVertical(String nt, Production p1, Production p2) {
		return nt + "[" + p1.getID().getLabel() + "] <--> " + nt + "[" + p2.getID().getLabel() + "]";
	}

	/**
	 * Checks whether the given string is horizontally ambiguous. 
	 */
	private boolean parseCheckHorizontal(Production p, int index, HorizontalOverlapString x) {
		Production p1 = new Production("$", p.getEntities().subList(0, index), false, p.getID(), 0);
		Production p2 = new Production("$", p.getEntities().subList(index, p.getEntities().size()), false, p.getID(), 0);
		boolean b = parser.check(x.getX(), p1) 
			&& parser.check(x.getXA(), p1) 
			&& parser.check(x.getY(), p2) 
			&& parser.check(x.getAY(), p2);
		if (debug)
			out.println("  parse check: " + (b ? "passed" : "failed"));
		return b;
	}
	
	/**
	 * Checks whether the given string is vertically ambiguous. 
	 */
	private boolean parseCheckVertical(Production p1, Production p2, VerticalOverlapString x) {
		boolean b =  parser.check(x.getString(), p1) && parser.check(x.getString(), p2);
		if (debug)
			out.println("  parse check: " + (b ? "passed" : "failed"));
		return b;
	}
	
	private String printEntities(Production p, int from, int to) {
		StringBuffer b = new StringBuffer();
		List<Entity> es = p.getEntities();
		for (int i = from; i < to; i++) {
			Entity e = es.get(i);
			b.append(' ').append(e.visitBy(new EntityVisitor<String>() {
				public String visitNonterminalEntity(NonterminalEntity e) {
					String s = e.getNonterminal();
					int i = s.lastIndexOf('#');
					if (g.isUnfolded() && i >= 0)
						return s.substring(0, i);
					return s;
				}
				public String visitRegexpTerminalEntity(RegexpTerminalEntity e) {
					return e.toString();
				}
				public String visitStringTerminalEntity(StringTerminalEntity e) {
					return "\"" + escape(decode(e.getString())) + "\"";
				}
				public String visitEOFTerminalEntity(EOFTerminalEntity e) {
					return e.toString();
				}
			}));
			if (e.isExplicitlyLabeled())
				b.append('[').append(e.getLabel()).append(']');
		}
		return b.toString();
	}
		
	private String decode(String s) {
		if (!g.isUnfolded())
			return s;
		StringBuilder b = new StringBuilder();
		for (int i = 0; i < s.length(); i += 2)
			b.append(s.charAt(i+1));
		return b.toString();
	}
	
	private void updateMaxMemory() {
		Runtime rt = Runtime.getRuntime();
		long mem = rt.totalMemory() - rt.freeMemory();
		if (mem > max_memory)
			max_memory = mem;
	}

	/**
	 * Returns maximal memory usage during last analysis.
	 * (Note: garbage collection might reduce this number.)
	 * @return memory usage
	 */
	public long getMaxMemory() {
		return max_memory;
	}
	
	/**
	 * Returns the number of potential horizontal ambiguities from last analysis.
	 * @return number of ambiguities
	 */
	public int getNumberOfPotentialHorizontalAmbiguities() {
		return potential_horizontal_ambiguities;
	}
	
	/**
	 * Returns the number of potential vertical ambiguities from last analysis.
	 * @return number of ambiguities
	 */
	public int getNumberOfPotentialVerticalAmbiguities() {
		return potential_vertical_ambiguities;
	}
	
	/**
	 * Returns the number of certain horizontal ambiguities from last analysis.
	 * @return number of ambiguities
	 */
	public int getNumberOfCertainHorizontalAmbiguities() {
		return certain_horizontal_ambiguities;
	}
	
	/**
	 * Returns the number of certain vertical ambiguities from last analysis.
	 * @return number of ambiguities
	 */
	public int getNumberOfCertainVerticalAmbiguities() {
		return certain_vertical_ambiguities;
	}
	
	/**
	 * Returns the number ambiguity checks that resulted in <code>OutOfMemoryError</code>.
	 * @return number of errors
	 */
	public int getNumberOfOutOfMemoryErrors() {
		return outofmemory;
	}
}
