package dk.brics.grammar.ambiguity;

import java.io.PrintWriter;

import dk.brics.grammar.Grammar;
import dk.brics.grammar.Production;

/**
 * Abstract base class for approximation strategies.
 */
abstract public class ApproximationStrategy {

	/**
	 * Current grammar.
	 */
	protected Grammar g;
	
	/**
	 * Verbose flag.
	 */
	protected boolean debug;
	
	/**
	 * Print stream for messages.
	 */
	protected PrintWriter out;
	
	private int passed_vertical_checks;

	private int failed_vertical_checks;
	
	private int na_vertical_checks;
	
	private int passed_horizontal_checks;
	
	private int failed_horizontal_checks;
	
	private int na_horizontal_checks;
	
	/**
	 * Value used by {@link #checkHorizontalOverlap(Production, int)} for "not applicable" responses.
	 */
	public static final HorizontalOverlapString HORIZONTAL_NOT_APPLICABLE = new HorizontalOverlapString("?", 0, 1);
	
	/**
	 * Value used by {@link #checkVerticalOverlap(Production, Production)} for "not applicable" responses.
	 */
	public static final VerticalOverlapString VERTICAL_NOT_APPLICABLE = new VerticalOverlapString("?");
	
	/**
	 * Returns the name of this approximation strategy.
	 * @return approximation strategy name
	 */
	abstract String getName();
	
	/**
	 * Abstract constructor.
	 */
	protected ApproximationStrategy() {}
	
	/** 
	 * Initializes for the given grammar.
	 * This method is invoked by <tt>AmbiguityAnalyzer</tt>.
	 * @param g current grammar
	 * @param out print writer for messages
	 * @param debug debug flag
	 */
	public final void init(Grammar g, PrintWriter out, boolean debug) {
		this.g = g;
		this.out = out;
		this.debug = debug;
		passed_horizontal_checks = failed_horizontal_checks = na_horizontal_checks = 
			passed_vertical_checks = failed_vertical_checks = na_vertical_checks = 0;
		init();
	}
	
	/** 
	 * Initializes for the current grammar.
	 * This method is typically implemented in a subclass.
	 * The default implementation does nothing.
	 * The grammar is available as {@link #g}.
	 */
	protected void init() {}
	
	/**
	 * Conservatively checks vertical overlap of two productions of the same nonterminal.
	 * This method is invoked by <tt>AmbiguityAnalyzer</tt>.
	 * @param p1 one production
	 * @param p2 another production
	 * @return potential example string in overlap, null if definitely no overlap,
	 *         {@link ApproximationStrategy#VERTICAL_NOT_APPLICABLE} can be used as a "don't know" value
	 */
	public final VerticalOverlapString verticalCheck(Production p1, Production p2) {
		VerticalOverlapString overlap = checkVerticalOverlap(p1, p2);
		if (overlap != ApproximationStrategy.VERTICAL_NOT_APPLICABLE)
			if (overlap == null)
				passed_vertical_checks++;
			else
				failed_vertical_checks++;
		else
			na_vertical_checks++;
		return overlap;
	}

	/**
	 * Conservatively checks horizontal overlap in a production.
	 * This method is invoked by <tt>AmbiguityAnalyzer</tt>.
	 * @param p a production
	 * @param index division point
	 * @return potential example string in overlap, null if definitely no overlap,
	 *         {@link ApproximationStrategy#HORIZONTAL_NOT_APPLICABLE} can be used as a "don't know" value
	 */
	public final HorizontalOverlapString horizontalCheck(Production p, int index) {
		HorizontalOverlapString overlap = checkHorizontalOverlap(p, index);
		if (overlap != ApproximationStrategy.HORIZONTAL_NOT_APPLICABLE)
			if (overlap == null)
				passed_horizontal_checks++;
			else
				failed_horizontal_checks++;
		else
			na_horizontal_checks++;
		return overlap;
	}

	/**
	 * Conservatively checks vertical overlap of two productions of the same nonterminal.
	 * This method is typically implemented in a subclass.
	 * The default implementation returns {@link ApproximationStrategy#VERTICAL_NOT_APPLICABLE}.
	 * @param p1 one production
	 * @param p2 another production
	 * @return potential example string in overlap, null if definitely no overlap,
	 *         {@link ApproximationStrategy#VERTICAL_NOT_APPLICABLE} can be used as a "don't know" value
	 */
	protected VerticalOverlapString checkVerticalOverlap(Production p1, Production p2) {
		return ApproximationStrategy.VERTICAL_NOT_APPLICABLE;
	}
	
	/**
	 * Conservatively checks horizontal overlap in a production.
	 * This method is typically implemented in a subclass.
	 * The default implementation returns {@link ApproximationStrategy#HORIZONTAL_NOT_APPLICABLE}.
	 * @param p a production
	 * @param index division point
	 * @return potential example string in overlap, null if definitely no overlap,
	 *         {@link ApproximationStrategy#HORIZONTAL_NOT_APPLICABLE} can be used as a "don't know" value
	 */
	protected HorizontalOverlapString checkHorizontalOverlap(Production p, int index) {
		return ApproximationStrategy.HORIZONTAL_NOT_APPLICABLE;
	}
	
	/**
	 * Invoked once per nonterminal when all its vertical checks are done (in the outer loop).
	 * Default implementation does nothing.
	 */
	protected void verticalDone() {}
	
	/**
	 * Invoked once per production when all its horizontal checks are done.
	 * Default implementation does nothing.
	 */
	protected void horizontalDone() {}
	
	/**
	 * Prints statistics from the last analysis.
	 * @param out print writer for output
	 */
	public void printStatistics(PrintWriter out) {
		out.println("  " + getName() + ": " 
					+ passed_vertical_checks + " passed, " 
					+ failed_vertical_checks + " failed, " 
					+ na_vertical_checks + " n/a vertical; " 
					+ passed_horizontal_checks + " passed, " 
					+ failed_horizontal_checks + " failed, " 
					+ na_horizontal_checks + " n/a horizontal");
	}
}
