package dk.brics.grammar.ambiguity;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import dk.brics.automaton.Automaton;
import dk.brics.grammar.EOFTerminalEntity;
import dk.brics.grammar.Entity;
import dk.brics.grammar.EntityVisitor;
import dk.brics.grammar.Grammar;
import dk.brics.grammar.NonterminalEntity;
import dk.brics.grammar.Production;
import dk.brics.grammar.ProductionID;
import dk.brics.grammar.RegexpTerminalEntity;
import dk.brics.grammar.StringTerminalEntity;
import dk.brics.grammar.operations.AutomataOperations;
import dk.brics.grammar.operations.Grammar2JSAGrammar;
import dk.brics.grammar.operations.ProductionLengthNormalizer;
import dk.brics.string.grammar.operations.Grammar2MLFA;
import dk.brics.string.mlfa.MLFA;
import dk.brics.string.mlfa.operations.MLFA2Automaton;

/**
 * Regular approximation based on Mohri &amp; Nederhof's technique.
 */
public class RegularApproximation extends ApproximationStrategy {
	
	/**
	 * Cache map from nonterminals to automata.
	 */
	private Map<String,Automaton> nt2a;
	
	/**
	 * Cache map from productions to automata.
	 */
	private Map<Production,Automaton> p2a;
	
	/**
	 * Grammar to MLFA converter.
	 */
	private Grammar2MLFA g2m;
	
	/**
	 * MLFA to automaton converter.
	 */
	private MLFA2Automaton m2a;
	
	/**
	 * Map from nonterminals to JSA grammar nonterminals.
	 */
	private Map<String,dk.brics.string.grammar.Nonterminal> nt_map;
	
	/**
	 * Constructs a new approximator.
	 */
	public RegularApproximation() {}
	
	/**
	 * Returns the name "regular".
	 */
	@Override
	public String getName() {
		return "regular";
	}

	@Override
	public void init() {
		System.out.println("POVODNA: \n" + g.toString());
		
//		/*
//		 * add new start symbol
//		 */
//		ArrayList<Entity> ent = new ArrayList<Entity>();
//		NonterminalEntity ne = new NonterminalEntity(g.getStart(), null, null);
//		ent.add(ne);
//		Production newStartProduction = new Production("START", ent, false, new ProductionID(), 0);
//		g.addProduction(newStartProduction, true);
//		g.setStart(newStartProduction.getNonterminal());
//		System.out.println("GRAMMAR WITH NEW START SYMBOL: \n" + g.toString());
		
		// normalize grammar
		Grammar normalized = new ProductionLengthNormalizer(out).normalize(g);
		System.out.println("NORMALIZOVANA: \n" + normalized.toString());
		if (debug) 
			out.print("normalized grammar:\n" + normalized.toString());
		// convert to JSA representation
		Grammar2JSAGrammar conv = new Grammar2JSAGrammar();
		dk.brics.string.grammar.Grammar jsagrammar = conv.convert(normalized);
		System.out.println("JSAg: \n" + jsagrammar.toString());
		nt_map = conv.getNonterminalMap();
		System.out.println("NT_MAP: " + nt_map.values() + "  / KEYS: " + nt_map.keySet());
		// run regular approximation and build automaton
		Collection<dk.brics.string.grammar.Nonterminal> hotspots = new ArrayList<dk.brics.string.grammar.Nonterminal>();
		for (String n : g.getNonterminals()) // use all nonterminals from the original grammar as hotspots
			hotspots.add(nt_map.get(n)); // TODO: make variant of RegularApproximation that approximates each time with just one hotspot
		int nonlinear = -1;
		if (debug) {
			nonlinear = jsagrammar.getNumberOfNonLinearComponents();
			out.println("components: " + jsagrammar.getNumberOfComponents() 
					+ ", nonlinear: " + nonlinear);
		}
		jsagrammar.approximateNonLinear(hotspots);
		if (debug && nonlinear > 0) {
			out.print("approximated grammar:\n" + jsagrammar.toString());
			for (String nt : g.getNonterminals())
				out.println(nt + " ~ " + nt_map.get(nt));
		}
		
		System.out.println("JSAG: \n :" + jsagrammar.toString());
		g2m = new Grammar2MLFA(jsagrammar);
		MLFA mlfa = g2m.convert();
		System.out.println("MLFA: " + mlfa.toString());
		
		if (debug)
			out.print("mlfa:\n" + mlfa.toString());
		
		m2a = new MLFA2Automaton(mlfa);
		System.out.println("M2A: \n" + m2a.toString() );
		nt2a = new HashMap<String,Automaton>();
		clearProductionCache();
	}
	
	@Override
	public void verticalDone() {
		clearProductionCache();
	}
	
	private void clearProductionCache() {
		p2a = new HashMap<Production,Automaton>();
	}

	private Automaton getAutomaton(String n) {
		Automaton a = nt2a.get(n);
		if (a == null) {
			System.out.println("NT_MAP.get(n) : " + nt_map.get(n).toString() + "  /  MLFAstatePAIR: " + g2m.getMLFAStatePair(nt_map.get(n)).toString());
			a = m2a.extract(g2m.getMLFAStatePair(nt_map.get(n)));
			nt2a.put(n, a);
		}
		return a;
	}
	
	private Automaton getAutomaton(Production p) {
		Automaton a = p2a.get(p);
		if (a == null) {
			a = getAutomaton(p.getEntities(), 0, p.getEntities().size());
			p2a.put(p, a);
		}
		return a;
	}

	private Automaton getAutomaton(List<Entity> entities, int from, int to) { // FIXME: handle unordered productions
		List<Automaton> as = new ArrayList<Automaton>();
		for (int i = from; i < to; i++) {
			Entity e = entities.get(i);
			Automaton a = e.visitBy(new EntityVisitor<Automaton>() {
				public Automaton visitNonterminalEntity(NonterminalEntity e) {
					return getAutomaton(e.getNonterminal());
				}
				public Automaton visitRegexpTerminalEntity(RegexpTerminalEntity e) { // FIXME: handle MAX regexps
					return e.getAutomaton();
				}
				public Automaton visitStringTerminalEntity(StringTerminalEntity e) {
					return Automaton.makeString(e.getString());
				}
				public Automaton visitEOFTerminalEntity(EOFTerminalEntity e) {
					return Automaton.makeEmptyString(); // FIXME: sound to treat EOF as "" - but not ideal!
				}
			});
			as.add(a);
		}		
		Automaton aut = Automaton.concatenate(as);
		//aut.minimize(); // TODO: need minimization here?
		return aut;
	}

	@Override
	public HorizontalOverlapString checkHorizontalOverlap(Production p, int index) { // FIXME: handle the case where exactly one of the two surrounding entities is ignorable
		Automaton a1 = getAutomaton(p.getEntities(), 0, index);
		Automaton a2 = getAutomaton(p.getEntities(), index, p.getEntities().size());
		Automaton overlap = AutomataOperations.getOverlap(a1, a2);
		// TODO: would it help to intersect overlap with p? (dangling-else...)
		return AutomataOperations.getOverlapString(overlap);
	}

	@Override
	public VerticalOverlapString checkVerticalOverlap(Production p1, Production p2) {
		Automaton a1 = getAutomaton(p1);
		System.out.println("A1: " + a1.toString());
		for (int i=0; i<6; i++) {
			System.out.println("A1 strings: " + a1.getStrings(i));
		}
		//System.out.println("A1 strings: " + a1.getFiniteStrings());
		
		Automaton a2 = getAutomaton(p2);
		System.out.println("A2: " + a2.toString());
		for (int i=0; i<6; i++) {
			System.out.println("A2 strings: " + a2.getStrings(i));
		}
		
		
		
		Automaton overlap = a1.intersection(a2); // TODO: make combined operation for intersection+getShortestExample
		// TODO: would it help to intersect overlap with p1/p2?
		String s = overlap.getShortestExample(true);
		System.out.println("String:" + s);
		
		return (s != null) ? new VerticalOverlapString(s) : null;
	}
}
