package dk.brics.grammar.ambiguity;

import java.util.ArrayList;
import java.util.List;

import dk.brics.automaton.Automaton;
import dk.brics.grammar.EOFTerminalEntity;
import dk.brics.grammar.Entity;
import dk.brics.grammar.EntityVisitor;
import dk.brics.grammar.Grammar;
import dk.brics.grammar.NonterminalEntity;
import dk.brics.grammar.Production;
import dk.brics.grammar.RegexpTerminalEntity;
import dk.brics.grammar.StringTerminalEntity;

/**
 * Variation of {@link RegularApproximation} that adds extra approximation on unfolded grammars.
 */
public class RegularApproximation2 extends RegularApproximation {
	
	boolean enabled;

	/**
	 * Constructs a new approximator.
	 */
	public RegularApproximation2() {}
	
	/**
	 * Returns the name "regular2".
	 */
	@Override
	public String getName() {
		return "regular2";
	}

	@Override
	public void init() {
		// hack to rediscover the unfold level
		int maxlevel = 0;
		for (Production p : g.getProductions()) {
			int level = getLevel(p.getNonterminal());
			if (level > maxlevel)
				maxlevel = level;
		}
		if (maxlevel > 0)
			enabled = true;
		if (enabled) {
			g = new Grammar(g); // don't mess up the original grammar
			Automaton maxlevel_sigmastar = Automaton.makeChar((char)maxlevel).concatenate(Automaton.makeAnyChar()).repeat();
			final Entity sigmastar_entity = new RegexpTerminalEntity(maxlevel_sigmastar, false, "[ANY]", null, null);
			final int f_maxlevel = maxlevel;
			// in productions of nonterminals with level maxlevel-1, 
			// replace nonterminal entities with level maxlevel with "level maxlevel Sigma*"
			for (Production p : g.getProductions())
				if (getLevel(p.getNonterminal()) == maxlevel-1) {
					List<Entity> newentities = new ArrayList<Entity>();
					for (Entity e : p.getEntities())
						newentities.add(e.visitBy(new EntityVisitor<Entity>() {
							@Override
							public Entity visitNonterminalEntity(NonterminalEntity e) {
								if (getLevel(e.getNonterminal()) == f_maxlevel)
									return sigmastar_entity;
								return e;
							}
							@Override
							public Entity visitRegexpTerminalEntity(RegexpTerminalEntity e) {
								return e;
							}
							@Override
							public Entity visitStringTerminalEntity(StringTerminalEntity e) {
								return e;
							}
							@Override
							public Entity visitEOFTerminalEntity(EOFTerminalEntity e) {
								return e;
							}
						}));
					p.setEntities(newentities);
				}
			if (debug)
				out.print("regular2 grammar:\n" + g);
			super.init();
		}
	}
	
	private int getLevel(String nt) {
		int i = nt.indexOf('#');
		if (i > 0)
			return Integer.parseInt(nt.substring(i+1));
		return 0;
	}

	@Override
	public HorizontalOverlapString checkHorizontalOverlap(Production p, int index) {
		if (enabled)
			return super.checkHorizontalOverlap(p, index);
		return HORIZONTAL_NOT_APPLICABLE;
	}

	@Override
	public VerticalOverlapString checkVerticalOverlap(Production p1, Production p2) {
		if (enabled)
			return super.checkVerticalOverlap(p1, p2);
		return VERTICAL_NOT_APPLICABLE;
	}
}
