package dk.brics.grammar.ambiguity;

import java.util.HashMap;
import java.util.Map;
import java.util.Set;

import dk.brics.automaton.Automaton;

import dk.brics.grammar.EOFTerminalEntity;
import dk.brics.grammar.Entity;
import dk.brics.grammar.EntityVisitor;
import dk.brics.grammar.NonterminalEntity;
import dk.brics.grammar.Production;
import dk.brics.grammar.RegexpTerminalEntity;
import dk.brics.grammar.StringTerminalEntity;
import dk.brics.grammar.operations.CharSet;
import dk.brics.grammar.operations.FirstLastFinder;
import dk.brics.grammar.operations.NullableFinder;
import dk.brics.grammar.operations.TerminalFinder;

/**
 * Terminal approximation, using terminals, nullables, and FIRST/LAST sets.
 */
public class TerminalApproximation extends ApproximationStrategy { // TODO: split this into several approximations?
	
	private TerminalFinder tf;
	
	private Map<Production,Automaton> prodmay;
	
	private Map<String,Automaton> nonterminalmay;
	
	private Map<Entity,CharSet> entityfirst;
	
	private Map<Entity,CharSet> entitylast;
	
	private Map<Integer,CharSet> leftmay;
	
	private Map<Integer,CharSet> rightmay;
	
	private Set<Production> nullable_productions;
	
	/**
	 * Constructs a new approximator.
	 */
	public TerminalApproximation() {}
	
	/**
	 * Returns the name "terminal".
	 */
	@Override
	public String getName() {
		return "terminal";
	}

	@Override
	public void init() {
		if (debug)
			out.println("computing nullables");
		nullable_productions = new NullableFinder(g).getNullableProductions();
		if (debug)
			out.println("computing terminal sets");
		tf = new TerminalFinder(g);
		FirstLastFinder ff = new FirstLastFinder(g, g.isUnfolded());
		entityfirst = ff.getEntityFirst();
		entitylast = ff.getEntityLast();
		clearProductionCache();
		clearLeftCache();
		nonterminalmay = new HashMap<String,Automaton>();
	}

	@Override
	public void verticalDone() {
		clearProductionCache();
	}
	
	@Override
	public void horizontalDone() {
		clearLeftCache();
	}
	
	private void clearProductionCache() {
		prodmay = new HashMap<Production,Automaton>();
	}
	
	private void clearLeftCache() {
		leftmay = new HashMap<Integer,CharSet>();
		rightmay = new HashMap<Integer,CharSet>();
	}

	private Automaton getMayAutomaton(Production p) {
		Automaton a = prodmay.get(p);
		if (a == null) {
			a = Automaton.union(tf.getMayTerminals(p).getCollection()).repeat();
			prodmay.put(p, a);
		}
		return a;
	}
	
	private Automaton getMayAutomaton(Entity e) {
		return e.visitBy(new EntityVisitor<Automaton>() {
			@Override
			public Automaton visitNonterminalEntity(NonterminalEntity e) {
				Automaton a = nonterminalmay.get(e.getNonterminal());
				if (a == null) {
					a = Automaton.union(tf.getMayTerminals(e.getNonterminal()).getCollection());
					nonterminalmay.put(e.getNonterminal(), a);
				}
				return a;
			}
			@Override
			public Automaton visitRegexpTerminalEntity(RegexpTerminalEntity e) {
				return e.getAutomaton();
			}
			@Override
			public Automaton visitStringTerminalEntity(StringTerminalEntity e) {
				return Automaton.makeString(e.getString());
			}
			@Override
			public Automaton visitEOFTerminalEntity(EOFTerminalEntity e) {
				return Automaton.makeEmptyString();
			}
		});
	}

	/** Finds may-chars for entities 0...(index-1) in p. **/
	private CharSet getMayCharsLeft(Production p, int index) {
		CharSet cs = leftmay.get(index);
		if (cs == null) {
			cs = new CharSet(getMayAutomaton(p.getEntities().get(index - 1)), false, false, g.isUnfolded());
			if (index > 1)
				cs.add(getMayCharsLeft(p, index - 1));
			leftmay.put(index, cs);
		}
		return cs;
	}
	
	/** Finds may-chars for entities index...(size-1) in p. **/
	private CharSet getMayCharsRight(Production p, int index) {
		CharSet cs = rightmay.get(index);
		if (cs == null) {
			cs = new CharSet(getMayAutomaton(p.getEntities().get(index)), false, false, g.isUnfolded());
			if (index + 1 < p.getEntities().size())
				cs.add(getMayCharsRight(p, index + 1));
			rightmay.put(index, cs);
		}
		return cs;
	}
	
	@Override
	public HorizontalOverlapString checkHorizontalOverlap(Production p, int index) {
		// check whether may-chars on left side are disjoint from FIRST chars of right side, or vice versa
		if (getMayCharsLeft(p, index).disjoint(entityfirst.get(p.getEntities().get(index))))
			return null;
		if (entitylast.get(p.getEntities().get(index - 1)).disjoint(getMayCharsRight(p, index)))
			return null;
		return HORIZONTAL_NOT_APPLICABLE;
	}

	@Override
	public VerticalOverlapString checkVerticalOverlap(Production p1, Production p2) { // FIXME: handle MAX regexps
		// check whether the empty string can/cannot be derived
		boolean p1_isempty = p1.getEntities().size() == 0;
		boolean p2_isempty = p2.getEntities().size() == 0;
		if (p1_isempty && p2_isempty)
			return new VerticalOverlapString("");
		boolean p1_acceptsempty = nullable_productions.contains(p1);
		boolean p2_acceptsempty = nullable_productions.contains(p2);
		if (p1_acceptsempty && p2_acceptsempty)
			return new VerticalOverlapString("");
		if ((p1_isempty && !p2_acceptsempty) || (p2_isempty && !p1_acceptsempty))
			return null;
		// if non-nullable, check whether the first FIRST sets or the last LAST sets of the productions are disjoint
		if (!p1_acceptsempty && !p2_acceptsempty) {
			if (entityfirst.get(p1.getEntities().get(0)).disjoint(entityfirst.get(p2.getEntities().get(0))))
				return null;
			if (entitylast.get(p1.getEntities().get(p1.getEntities().size() - 1)).disjoint(entitylast.get(p2.getEntities().get(p2.getEntities().size() - 1))))
				return null;
		}
		// TODO: check first+last *terminals* disjoint? (instead?)
		// check whether must-terminals from one production are disjoint from may-terminals from the other
		for (Automaton a1 : tf.getMustTerminals(p1)) 
			if (a1.intersection(getMayAutomaton(p2)).isEmpty())
				return null;
		for (Automaton a2 : tf.getMustTerminals(p2)) 
			if (a2.intersection(getMayAutomaton(p1)).isEmpty())
				return null;
		return VERTICAL_NOT_APPLICABLE;
	}
}
