package dk.brics.grammar.ast;

import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.LinkedHashMap;

import dk.brics.grammar.ProductionID;

/**
 * AST branch node.
 */
public class BranchNode extends Node {
	
	/** Node name. */
	private String name;

	/** Map from edge names to children, null if empty. */
	private Map<String, Node> map;

	/** ID of the corresponding grammar production. */ 
	private ProductionID id;

	/**
	 * Constructs a new AST node with no edges (yet).
	 * @param id production ID
	 * @param name node name
	 * @param i from index
	 * @param j to index (not included)
	 */
	public BranchNode(ProductionID id, String name, int i, int j) {
		super(i, j);
		this.id = id;
		this.name = name;
	}
	
	/**
	 * Constructs a new AST node from an existing one by adding an edge.
	 * @param n existing node
	 * @param edge edge name
	 * @param m child node
	 */
	public BranchNode(BranchNode n, String edge, Node m) {
		super(n.getFromIndex(), m.getToIndex());
		id = n.id;
		name = n.name;
		if (n.map == null)
			map = new LinkedHashMap<String, Node>();
		else
			map = new LinkedHashMap<String, Node>(n.map);
		map.put(edge, m);
	}

	/**
	 * Constructs a new AST node from an existing one by updating 'to'.
	 * @param n existing node
	 * @param to to index
	 */
	public BranchNode(BranchNode n, int to) {
		super(n.getFromIndex(), to);
		id = n.id;
		name = n.name;
		map = n.map; // note: sharing map
	}
	
	/**
	 * Returns a child node.
	 * @param name name of child node
	 * @return node, null if absent
	 */
	public Node getChild(String name) {
		if (map == null)
			return null;
		return map.get(name);
	}
	
	/**
	 * Returns a branch child node.
	 * @param name name of branch child node
	 * @return node, null if absent
	 * @throws ClassCastException if the child is not a branch node 
	 */
	public BranchNode getBranchChild(String name) throws ClassCastException {
		if (map == null)
			return null;
		return (BranchNode)map.get(name);
	}
	
	/**
	 * Returns the production ID.
	 * @return production ID
	 */
	public ProductionID getProductionID() {
		return id;
	}
	
	/**
	 * Returns the node name.
	 * @return name
	 */
	public String getName() {
		return name;
	}
	
	/**
	 * Returns the set of child names in this node.
	 * @return set of child names
	 */
	public Set<String> getChildNames() {
		if (map == null)
			return new HashSet<String>();
		return map.keySet();
	}
	
	/**
	 * Replaces an child with the children of another node.
	 * @param name child to replace
	 * @param n node containing children to copy
	 */
	public void replaceChild(String name, BranchNode n) {
		map.remove(name);
		if (n.map != null)
			map.putAll(n.map);
	}
	
	/**
	 * Returns the substring of the leaf child of the given name.
	 * @throws ClassCastException if the child is not a leaf node 
	 */
	public String getLeafString(String name, AST t) throws ClassCastException {
		LeafNode s = (LeafNode)getChild(name);
		if (s == null)
			return null;
		return s.getString(t.getOriginalString());
	}
	
	/**
	 * Returns the production label of this node.
	 * @return label
	 */
	public String getLabel() {
		return getProductionID().getLabel();
	}
	
	@Override
	public void print(String x, StringBuilder b) {
		b.append(name);
		if (id.getLabel() != null)
			b.append("{" + id.getLabel() + "}");
		b.append("(").append(getFromIndex()).append('-').append(getToIndex()).append(")[");
		boolean first = true;
		if (map != null)
			for (String m : map.keySet()) {
				if (!first)
					b.append(",");
				b.append(m).append("->");
				map.get(m).print(x, b);
				first = false;
			}
		b.append("]");
	}
	
	@Override 
	void traverse(NodeVisitor v) {
		if (map != null)
			for (Map.Entry<String,Node> e : map.entrySet())
				e.getValue().traverse(v);
		v.visitBranchNode(this);
	}

	@Override 
	public void visitBy(NodeVisitor v) {
		v.visitBranchNode(this);
	}
}
