package dk.brics.grammar.main;

import java.io.ByteArrayOutputStream;
import java.io.PrintWriter;

import dk.brics.grammar.Grammar;
import dk.brics.grammar.GrammarException;
import dk.brics.grammar.ambiguity.AmbiguityAnalyzer;
import dk.brics.grammar.ast.AST;
import dk.brics.grammar.ast.DotPrinter;
import dk.brics.grammar.operations.GrammarTokenizer;
import dk.brics.grammar.operations.Unfolder;
import dk.brics.grammar.parser.ParseException;
import dk.brics.grammar.parser.Parser;
import dk.brics.grammar.parser.String2Grammar;

/**
 * Parser and grammar ambiguity checker.
 * <p>
 * This tool can be used to
 * <ul>
 * <li>check that a text is syntactically correct according to the a grammar,
 * <li>check that a grammar itself is syntactically correct,
 * <li>print parse trees (Graphviz dot format), and
 * <li>analyze grammars for potential ambiguity (see {@link AmbiguityAnalyzer}).
 * </ul> 
 * @see MainCommandLine
 * @see MainGUI
 * @see MainServlet
 */
public class Main {
	
	private Main() {}
	
	/**
	 * Parser and grammar ambiguity checker.
	 * If no arguments are provided, the GUI is started; otherwise, the program runs in command-line mode. 
	 * Run with <code>-h</code> to see the command-line usage.
	 * @param args arguments
	 * @see #run(String, String, String, String, boolean, int, String, String, boolean, boolean, boolean, boolean, PrintWriter)
	 */
	public static void main(String[] args) {
		if (args.length == 0)
			MainGUI.main(args);
		else 
			MainCommandLine.main(args);
	}

	/**
	 * Front-end for parser and grammar ambiguity checker.
	 * @param grammar grammar
	 * @param grammar_location location of grammar (typically a directory path or a URL) 
	 * @param text text to parse (null if none)
	 * @param text_location location of text (null if none)
	 * @param ambiguity_analysis if true, run ambiguity analysis
	 * @param unfold_level unfold number
	 * @param unfold_left unfolding left parentheses symbols
	 * @param unfold_right unfolding right parentheses symbols
	 * @param verbose if true, verbose output 
	 * @param extrastats if true, print extra statistics
	 * @param dump_ast if true, dump AST after parsing
	 * @param tokenize if true, tokenize grammar before ambiguity analysis
	 * @param out print writer for output
	 * @return status code (0 = everything ok, -1 = illegal input, -2 = parse error in grammar, 
	 *                      -3 = parse error in text, -4 = other error in grammar, 
	 *                      -5 = potential ambiguity in grammar)
	 * @throws ClassNotFoundException if an approximation strategy class was not found
	 * @throws IllegalAccessException if an approximation strategy class or its nullary constructor is not accessible
	 * @throws InstantiationException if an approximation strategy class cannot be instantiated
	 * @throws IllegalArgumentException if the grammar is not balanced with the given unfolding parentheses
	 */
	public static int run(String grammar, String grammar_location, String text, String text_location,
			boolean ambiguity_analysis, int unfold_level, String unfold_left, String unfold_right, 
			boolean verbose, boolean dump_ast, boolean tokenize, boolean extrastats, 
			PrintWriter out) 
	throws InstantiationException, IllegalAccessException, ClassNotFoundException, IllegalArgumentException {
		AmbiguityAnalyzer a = null;
		Grammar g = null;
		Parser p = null;
		long begin = System.currentTimeMillis();
		long time_analysis = -1;
		long time_text = -1;
		try {
			try {
				if (verbose)
					out.println("loading grammar " + grammar_location);
				ByteArrayOutputStream b = new ByteArrayOutputStream();
				PrintWriter bout = new PrintWriter(b, true);
				g = new String2Grammar().convert(grammar, grammar_location, bout);
				bout.close();
				out.print(b);
				if (!ambiguity_analysis && text_location == null) {
					if (b.size() == 0)
						out.println("no errors found in the grammar!"); 
					else
						out.println("errors found in the grammar!"); 
				}
			} catch (ParseException e) {
				out.println("*** " + grammar_location + ": " + e.getMessage());
				printStatus(begin, time_analysis, time_text, p, a, false, out);
				return -2;
			}
			if (ambiguity_analysis) {
				Grammar g2 = g;
				if (tokenize) {
					if (verbose)
						out.println("tokenizing grammar");
					new GrammarTokenizer().tokenize(g);
				}
				if (unfold_level > 0) {
					if (verbose)
						out.println("unfolding grammar");
					g2 = new Unfolder(out).unfold(g2, unfold_level, unfold_left, unfold_right);
				}
				a = new AmbiguityAnalyzer(out, verbose);
				long t = System.currentTimeMillis();
				a.analyze(g2);
				time_analysis = System.currentTimeMillis() - t;
			}
			if (text_location != null) {
				try {
					p = new Parser(g, out);
					long t = System.currentTimeMillis();
					if (verbose)
						out.println("parsing text...");
					AST ast = p.parse(text, text_location);
					time_text = System.currentTimeMillis() - t;
					if (dump_ast) 
						new DotPrinter(out).print(ast);
					else
						out.println("no syntax errors!");
				} catch (ParseException e) {
					out.println("*** " + text_location + ": " + e.getMessage());
					out.println("syntax error found!");
					if (verbose) {
						out.println();
						printStatus(begin, time_analysis, time_text, p, a, false, out);
						return -3;
					}
				}
			} 
		} catch (GrammarException e) {
			out.println("*** error in grammar: " + e.getMessage());
			return -4;
		}
		if (!dump_ast || ambiguity_analysis) {
			boolean amb = false;
			if (ambiguity_analysis) {
				int pv = a.getNumberOfPotentialVerticalAmbiguities();
				int ph = a.getNumberOfPotentialHorizontalAmbiguities();
				int cv = a.getNumberOfCertainVerticalAmbiguities();
				int ch = a.getNumberOfCertainHorizontalAmbiguities();
				int err = a.getNumberOfOutOfMemoryErrors();
				if (pv + ph + cv + ch + err == 0)
					out.println("the grammar is unambiguous!");
				else {
					if (cv + ch > 0) 
						out.println("the grammar is ambiguous!");
					else
						out.println("the grammar might be ambiguous, but I'm not sure...");
					amb = true;
				}
			}
			if (verbose) {
				out.println();
				if (ambiguity_analysis)
					a.printStatistics(out);
				printStatus(begin, time_analysis, time_text, p, a, extrastats, out);
			}
			if (amb)
				return -5;
		}
		out.flush();
		return 0;
	}
	
	private static void printStatus(long begin, long time_analysis, long time_text, Parser p, AmbiguityAnalyzer a, boolean extrastats, PrintWriter out) {
	    if (a != null && extrastats)
	    	out.println("maximal jvm memory usage during ambiguity analysis (without intensive gc): " + a.getMaxMemory() + " bytes");
	    if (time_analysis != -1)
	    	out.println("time for analyzing ambiguity: " + time_analysis + "ms");
	    if (time_text != -1)
	    	out.println("time for parsing input text: " + time_text + "ms");
	    out.println("total time: " + (System.currentTimeMillis() - begin) + "ms");
	    if (extrastats && p != null) {
	    	out.println("maximal jvm memory usage during parsing (without intensive gc): " + p.getMaxMemory() + " bytes");
	    	out.println("total number of parse states: " + p.getTotalStates());
	    	out.println("maximal number of pending parse states: " + p.getMaxStates());
	    }
	}
}
