package dk.brics.grammar.main;

import java.io.IOException;
import java.io.PrintWriter;
import java.nio.charset.Charset;

import dk.brics.misc.Loader;

/**
 * Command-line interface to the parser and grammar ambiguity checker.
 */
public class MainCommandLine {

	private MainCommandLine() {}
	
	/**
	 * Runs the command-line interface to the parser and grammar ambiguity checker.
	 */
	public static void main(String[] args) {
		String grammar_encoding, text_encoding;
		grammar_encoding = text_encoding = Charset.defaultCharset().name();
		boolean dump_ast = false, ambiguity_analysis = false, verbose = false, tokenize = false;;
		String grammar_location = null;
		String text_location = null;
		int unfold_level = 0;
		String unfold_left = "";
		String unfold_right = "";
		boolean bad = false;
		boolean help = false;
		for (int i = 0; i < args.length; i++) {
			String a = args[i];
			if (a.startsWith("-")) {
				if (a.equals("-d"))
					dump_ast = true;
				else if (a.equals("-a"))
					ambiguity_analysis = true;
				else if (a.equals("-v"))
					verbose = true;
				else if (a.equals("-z"))
					tokenize = true;
				else if (a.equals("-h"))
					help = bad = true;
				else if (a.equals("-g") && i + 1 < args.length)
					grammar_encoding = args[++i];
				else if (a.equals("-t") && i + 1 < args.length)
					text_encoding = args[++i];
				else if (a.equals("-u") && i + 1 < args.length) {
					try {
						unfold_level = Integer.parseInt(args[++i]);
					} catch (NumberFormatException e) {
						bad = true;
					}
					if (unfold_level < 0)
						bad = true;
				} else if (a.equals("-l") && i + 1 < args.length)
					unfold_left = args[++i];
				else if (a.equals("-r") && i + 1 < args.length)
					unfold_right = args[++i];
				else
					bad = true;
			} else {
				if (grammar_location == null)
					grammar_location = a;
				else if (text_location == null)
					text_location = a;
				else
					bad = true;
			}
		}
		if (grammar_location == null && !help) {
			bad = true;
			System.out.println("*** error: no grammar specified");
		}
		if (bad) {
			System.out.print("Usage: java dk.brics.grammar.main.Main [Options...] <path or URL of grammar> [ <path or URL of text to parse> ]\n" +
					"\n" +
					"This tool checks whether the given text is syntactically correct according to the\n" +
					"given grammar and that the grammar itself is syntactically correct.\n" +
					"It can also check whether the grammar is ambiguous using the technique described in\n" +
					"\"Analyzing Ambiguity of Context-Free Grammars\", Claus Brabrand, Robert Giegerich,\n"+
					"and Anders Mller, CIAA 2007.\n" +
					"If only a grammar is given, only the grammar checks are performed.\n" +	 
					"\n" +
					"Options:\n" +
					"-a                     analyze the grammar for potential ambiguity\n" +
					"-v                     verbose, print progress information and statistics\n" +
					"-d                     dump AST in Graphviz dot format after parsing\n" +
					"-g <encoding>          character encoding used by the grammar\n" +
					"-t <encoding>          character encoding used by the text\n" +
					"-u <unfold level>      grammar unfolding level (for ambiguity analysis)\n" +
					"-l <left parentheses>  left parentheses symbols for grammar unfolding\n" +
					"-r <right parentheses> right parentheses symbols for grammar unfolding\n" +
					"-z                     tokenize grammar (for ambiguity analysis)\n" +
					"\n" +
					"System properties:\n" +
					"-Ddk.brics.grammar.parser.debug            extra output from parser\n" +
					"-Ddk.brics.grammar.ambiguity.debug         extra output from ambiguity analyzer\n" +
					"-Ddk.brics.grammar.ambiguity.noparsecheck  omit parse check in ambiguity analyzer\n" +
					"-Ddk.brics.grammar.ambiguity.ignorables    force enabling ignorables mode in ambiguity analyzer\n" +
					"-Ddk.brics.grammar.ambiguity.strategies=[comma separated list of ApproximationStrategy classes]\n");
			System.exit(-1);
		}
		try {
			String grammar = Loader.getString(grammar_location, grammar_encoding);
			String text = text_location != null ? Loader.getString(text_location, text_encoding) : null;
			int status = Main.run(grammar, grammar_location, text, text_location, ambiguity_analysis, 
					unfold_level, unfold_left, unfold_right, verbose, dump_ast, tokenize, true, new PrintWriter(System.out, true));
			if (status != 0)
				System.exit(status);
		} catch (IllegalArgumentException e) {
			fail(e);
		} catch (IOException e) {
			fail(e);
		} catch (InstantiationException e) {
			fail(e);
		} catch (IllegalAccessException e) {
			fail(e);
		} catch (ClassNotFoundException e) {
			fail(e);
		}
	}

	private static void fail(Exception e) {
		System.out.println("*** error: " + e.getMessage());
		System.exit(-1);
	}
}
