package dk.brics.grammar.main;

import static java.awt.EventQueue.invokeLater;

import java.awt.*;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.Writer;
import java.nio.charset.Charset;

import javax.swing.*;
import javax.swing.filechooser.FileFilter;

import dk.brics.misc.Loader;

/**
 * Graphical interface to the parser and grammar ambiguity checker.
 * <p>
 * To use Graphviz dot, set the system property <tt>dk.brics.grammar.dot</tt> to the path to the <tt>dot</tt> executable.  
 */
public class MainGUI {
	
	private final static String[] encodings = {
			"default",
			"utf-8 (Unicode, worldwide)",
			"utf-16 (Unicode, worldwide)",
			"iso-8859-1 (Western Europe)",
			"iso-8859-2 (Central Europe)",
			"iso-8859-3 (Southern Europe)",
			"iso-8859-4 (North European)",
			"iso-8859-5 (Cyrillic)",
			"iso-8859-6-i (Arabic)",
			"iso-8859-7 (Greek)",
			"iso-8859-8 (Hebrew, visual)",
			"iso-8859-8-i (Hebrew, logical)",
			"iso-8859-9 (Turkish)",
			"iso-8859-10 (Latin 6)",
			"iso-8859-11 (Latin/Thai)",
			"iso-8859-13 (Latin 7, Baltic Rim)",
			"iso-8859-14 (Latin 8, Celtic)",
			"iso-8859-15 (Latin 9)",
			"iso-8859-16 (Latin 10)",
			"us-ascii (basic English)",
			"euc-jp (Japanese, Unix)",
			"shift_jis (Japanese, Win/Mac)",
			"iso-2022-jp (Japanese, email)",
			"euc-kr (Korean)",
			"ksc_5601 (Korean)",
			"gb2312 (Chinese, simplified)",
			"gb18030 (Chinese, simplified)",
			"big5 (Chinese, traditional)",
			"Big5-HKSCS (Chinese, Hong Kong)",
			"tis-620 (Thai)",
			"koi8-r (Russian)",
			"koi8-u (Ukrainian)",
			"iso-ir-111 (Cyrillic KOI-8)",
			"macintosh (MacRoman)",
			"windows-1250 (Central Europe)",
			"windows-1251 (Cyrillic)",
			"windows-1252 (Western Europe)",
			"windows-1253 (Greek)",
			"windows-1254 (Turkish)",
			"windows-1255 (Hebrew)",
			"windows-1256 (Arabic)",
			"windows-1257 (Baltic Rim)"
	};
	
	private MainGUI() {}

	/**
	 * Runs the graphical interface to the parser and grammar ambiguity checker.
	 */
	public static void main(String[] args) {
		invokeLater(new Runnable() {
            public void run() {
            	try {
					UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
	                new Window().setVisible(true);
				} catch (Exception e) {
					System.err.println(e.getMessage());
				}
            }
        });
	}
	
	static class Window extends JFrame {

		JTextField grammarfile_textfield;
		JComboBox grammarencoding_combobox;
		JTextField textfile_textfield;
		JComboBox textencoding_combobox;
		JCheckBox verbose_checkbox; 	
		JCheckBox dumpast_checkbox; 	
		JCheckBox tokenize_checkbox; 	
		JTextField unfoldlevel_textfield;
		JTextField unfoldleft_textfield;
		JTextField unfoldright_textfield;
		JTextArea output_textarea;
		JFileChooser grammarfile_chooser;
		JFileChooser textfile_chooser;
		JButton check_button;
		JButton parse_button;
		JButton analyze_button;
		JButton abort_button;
		JButton graphviz_button;
		PrintWriter out;
		Process dotproc;
		boolean graphviz_aborted;
		
		Window() {
			JLabel grammarfile_label = new JLabel("Grammar URL or file:");
			grammarfile_textfield = new JTextField(40);
			JButton grammarfile_button = new JButton("Open file...");
			JLabel grammarencoding_label = new JLabel("Encoding:");
			grammarencoding_combobox = new JComboBox(encodings);

			JLabel textfile_label = new JLabel("Text URL or file:");
			textfile_textfield = new JTextField(40);
			JButton textfile_button = new JButton("Open file...");
			JLabel textencoding_label = new JLabel("Encoding:");
			textencoding_combobox = new JComboBox(encodings);
			
			check_button = new JButton("Check grammar");
			parse_button = new JButton("Parse text");
			analyze_button = new JButton("Analyze grammar ambiguity");
			abort_button = new JButton("Abort");
			graphviz_button = new JButton("Run Graphviz");
			JButton about_button = new JButton("About");
			JButton exit_button = new JButton("Exit");

			verbose_checkbox = new JCheckBox("Verbose"); 
			dumpast_checkbox = new JCheckBox("Dump AST"); 
			tokenize_checkbox = new JCheckBox("Tokenize grammar"); 
			
			JLabel unfoldlevel_label = new JLabel("Unfold level:");
			unfoldlevel_textfield = new JTextField(1);
			unfoldlevel_textfield.setText("0");
			JLabel unfoldleft_label = new JLabel("Left parentheses:");
			unfoldleft_textfield = new JTextField(3);
			JLabel unfoldright_label = new JLabel("Right parentheses:");
			unfoldright_textfield = new JTextField(3);

			output_textarea = new JTextArea();
			output_textarea.setEditable(false);

			grammarfile_chooser = new JFileChooser();
			grammarfile_chooser.setFileFilter(new FileFilter() {
				@Override
				public boolean accept(File f) {
					return f.isDirectory() || f.getName().endsWith(".cfg");
				}
				@Override
				public String getDescription() {
					return "grammar files (.cfg)";
				}
			});
			textfile_chooser = new JFileChooser();

			grammarfile_label.setLabelFor(grammarfile_textfield);
			grammarencoding_label.setLabelFor(grammarencoding_combobox);
			textfile_label.setLabelFor(textfile_textfield);
			textencoding_label.setLabelFor(textencoding_combobox);

			JPanel inputfile_panel = new JPanel();
	        inputfile_panel.setLayout(new GridBagLayout());
	        GridBagConstraints c = new GridBagConstraints();
	        c.fill = GridBagConstraints.NONE;
	        c.gridx = 0;
	        c.gridy = 0;
	        c.anchor = GridBagConstraints.WEST;
	        inputfile_panel.add(grammarfile_label, c);
	        c.fill = GridBagConstraints.HORIZONTAL;
	        c.gridx = 1;
	        c.weightx = 1;
	        inputfile_panel.add(grammarfile_textfield, c);
	        c.fill = GridBagConstraints.NONE;
	        c.gridx = 2;
	        c.weightx = 0;
	        inputfile_panel.add(grammarfile_button, c);
	        c.fill = GridBagConstraints.NONE;
	        c.gridx = 3;
	        c.insets = new Insets(0, 20, 0, 0);
	        inputfile_panel.add(grammarencoding_label, c);
	        c.fill = GridBagConstraints.NONE;
	        c.gridx = 4;
	        c.insets = new Insets(0, 0, 0, 0);
	        inputfile_panel.add(grammarencoding_combobox, c);
	        c.fill = GridBagConstraints.NONE;
	        c.gridx = 0;
	        c.gridy = 1;
	        inputfile_panel.add(textfile_label, c);
	        c.fill = GridBagConstraints.HORIZONTAL;
	        c.gridx = 1;
	        c.weightx = 1;
	        inputfile_panel.add(textfile_textfield, c);
	        c.fill = GridBagConstraints.NONE;
	        c.gridx = 2;
	        c.weightx = 0;
	        inputfile_panel.add(textfile_button, c);
	        c.fill = GridBagConstraints.NONE;
	        c.gridx = 3;
	        c.insets = new Insets(0, 20, 0, 0);
	        inputfile_panel.add(textencoding_label, c);
	        c.fill = GridBagConstraints.NONE;
	        c.gridx = 4;
	        c.insets = new Insets(0, 0, 0, 0);
	        inputfile_panel.add(textencoding_combobox, c);
	        inputfile_panel.setBorder(
	                BorderFactory.createCompoundBorder(
	                                BorderFactory.createTitledBorder("Input files"),
	                                BorderFactory.createEmptyBorder(5,5,5,5)));
	        
			JPanel option_panel = new JPanel();
			option_panel.setLayout(new GridBagLayout());
	        c.fill = GridBagConstraints.NONE;
	        c.gridx = 0;
	        c.gridy = 0;
	        c.insets = new Insets(0, 0, 0, 0);
	        option_panel.add(verbose_checkbox, c);
	        c.gridx = 1;
	        option_panel.add(dumpast_checkbox, c);
	        c.gridx = 2;
	        c.insets = new Insets(0, 0, 0, 30);
	        option_panel.add(tokenize_checkbox, c);
	        c.gridx = 3;
	        c.insets = new Insets(0, 0, 0, 0);
	        option_panel.add(unfoldlevel_label, c);
	        c.gridx = 4;
	        c.insets = new Insets(0, 0, 0, 10);
	        option_panel.add(unfoldlevel_textfield, c);
	        c.gridx = 5;
	        c.insets = new Insets(0, 0, 0, 0);
	        option_panel.add(unfoldleft_label, c);
	        c.gridx = 6;
	        c.insets = new Insets(0, 0, 0, 10);
	        option_panel.add(unfoldleft_textfield, c);
	        c.gridx = 7;
	        c.insets = new Insets(0, 0, 0, 0);
	        option_panel.add(unfoldright_label, c);
	        c.gridx = 8;
	        option_panel.add(unfoldright_textfield, c);
			option_panel.setBorder(BorderFactory.createEmptyBorder(5,5,5,5));
        
			JPanel command1_panel = new JPanel();
	        command1_panel.setLayout(new GridBagLayout());
	        c.anchor = GridBagConstraints.CENTER;
	        c.gridx = 0;
	        command1_panel.add(check_button, c);
	        c.gridx = 1;
	        c.insets = new Insets(0, 5, 0, 0);
	        command1_panel.add(parse_button, c);
	        c.gridx = 2;
	        command1_panel.add(analyze_button, c);
	        c.insets = new Insets(0, 0, 0, 0);
	        command1_panel.setBorder(BorderFactory.createEmptyBorder(5,5,5,5));

			JPanel command2_panel = new JPanel();
	        command2_panel.setLayout(new GridBagLayout());
	        c.anchor = GridBagConstraints.CENTER;
	        c.gridx = 0;
	        command2_panel.add(abort_button, c);
	        c.gridx = 1;
	        c.insets = new Insets(0, 40, 0, 0);
	        command2_panel.add(graphviz_button, c);
	        c.gridx = 2;
	        c.insets = new Insets(0, 50, 0, 0);
	        command2_panel.add(about_button, c);
	        c.gridx = 3;
	        c.insets = new Insets(0, 20, 0, 0);
	        command2_panel.add(exit_button, c);
	        c.insets = new Insets(0, 0, 0, 0);
	        command2_panel.setBorder(BorderFactory.createEmptyBorder(5,5,5,5));
	        abort_button.setEnabled(false);
	        graphviz_button.setEnabled(false);

			JScrollPane output_pane = new JScrollPane(output_textarea);
	        output_pane.setVerticalScrollBarPolicy(JScrollPane.VERTICAL_SCROLLBAR_ALWAYS);
	        output_pane.setPreferredSize(new Dimension(400, 250));
	        output_pane.setBorder(
	            BorderFactory.createCompoundBorder(
	                BorderFactory.createCompoundBorder(
	                                BorderFactory.createTitledBorder("Output"),
	                                BorderFactory.createEmptyBorder(5,5,5,5)),
	                output_pane.getBorder()));
			out = new PrintWriter(new TextAreaWriter(output_textarea)); 
	        
			setTitle("dk.brics.grammar");
			setMinimumSize(new Dimension(700, 300));
			getContentPane().setLayout(new BoxLayout(getContentPane(), BoxLayout.Y_AXIS));
			add(inputfile_panel);
	        add(option_panel);
	        add(command1_panel);
	        add(command2_panel);
	        add(output_pane);
	        inputfile_panel.setMaximumSize(new Dimension(Integer.MAX_VALUE, inputfile_panel.getHeight()));
	        option_panel.setMaximumSize(new Dimension(Integer.MAX_VALUE, option_panel.getHeight()));
	        command1_panel.setMaximumSize(new Dimension(Integer.MAX_VALUE, command1_panel.getHeight()));
	        command2_panel.setMaximumSize(new Dimension(Integer.MAX_VALUE, command2_panel.getHeight()));

	        pack();

	        final Component frame = this;
			setDefaultCloseOperation(javax.swing.WindowConstants.DO_NOTHING_ON_CLOSE);
	        addWindowListener(new WindowAdapter() {
	        	@Override
				public void windowClosing(WindowEvent e) {
	        		exit();
	        	}
	        });
	        about_button.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					JOptionPane.showMessageDialog(frame, 
							"Copyright (C) 2005-2008 Anders Mller\n\n" +
							"Please see http://www.brics.dk/grammar/ for information about this software.", 
							"dk.brics.grammar", JOptionPane.INFORMATION_MESSAGE);
				}
			});
	        exit_button.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					exit();
				}
			});
	        grammarfile_button.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					if (grammarfile_chooser.showOpenDialog(frame) == JFileChooser.APPROVE_OPTION)
						grammarfile_textfield.setText(grammarfile_chooser.getSelectedFile().getPath());				
				}    	
	        });
	        textfile_button.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					if (textfile_chooser.showOpenDialog(frame) == JFileChooser.APPROVE_OPTION)
						textfile_textfield.setText(textfile_chooser.getSelectedFile().getPath());				
				}    	
	        });
	        check_button.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					action("check");
				}
	        });
	        parse_button.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					action("parse");					
				}
	        });
	        analyze_button.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					action("analyze");
				}
	        });
	        graphviz_button.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					graphviz();
				}
	        });
	        verbose_checkbox.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					synchronized (frame) {
						dumpast_checkbox.setSelected(false);
					}
				}
	        });
	        dumpast_checkbox.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					synchronized (frame) {
						verbose_checkbox.setSelected(false);
					}
				}
	        });
	     
	        grammarfile_textfield.setToolTipText("location of the context-free grammar");
	        grammarfile_label.setToolTipText(grammarfile_textfield.getToolTipText());
	        grammarfile_button.setToolTipText("select a grammar file");
	        textfile_textfield.setToolTipText("location of the text file to parse");
	        textfile_label.setToolTipText(textfile_textfield.getToolTipText());
	        textfile_button.setToolTipText("select a text file to parse");
	        grammarencoding_combobox.setToolTipText("character encoding of grammar file");
	        grammarencoding_label.setToolTipText(grammarencoding_combobox.getToolTipText());
	        textencoding_combobox.setToolTipText("character encoding of text file");
	        textencoding_label.setToolTipText(textencoding_combobox.getToolTipText());
	        verbose_checkbox.setToolTipText("if checked, verbose output is generated");
	        dumpast_checkbox.setToolTipText("if checked, the abstract syntax tree is output after successful parsing");
	        tokenize_checkbox.setToolTipText("if checked, the ambiguity analyzer may assume that all terminals are maximal match tokens");
	        unfoldlevel_textfield.setToolTipText("number of times to unfold grammar (default 0)");
	        unfoldlevel_label.setToolTipText(unfoldlevel_textfield.getToolTipText());
	        unfoldleft_textfield.setToolTipText("left parenthesis symbol for unfolding");
	        unfoldleft_label.setToolTipText(unfoldleft_textfield.getToolTipText());
	        unfoldright_textfield.setToolTipText("right parenthesis symbol for unfolding");
	        unfoldright_label.setToolTipText(unfoldright_textfield.getToolTipText());
	        check_button.setToolTipText("check that the grammar is syntactically correct");
	        parse_button.setToolTipText("try to parse the text according to the grammar");
	        analyze_button.setToolTipText("analyze the grammar for ambiguity");
	        abort_button.setToolTipText("abort current action");
	        graphviz_button.setToolTipText("run Graphviz to show the AST graphically (after parsing with 'Dump AST' enabled')");
	        about_button.setToolTipText("information about this tool");
	    }
		
		@Override
		public void setVisible(boolean b) {
		  if (b) {
			  Dimension screensize = Toolkit.getDefaultToolkit().getScreenSize();
			  setLocation((screensize.width-getWidth())/2, (screensize.height-getHeight())/2);
		  }
		  super.setVisible(b);
		}

		private synchronized void action(final String action) {
			new SwingWorker<Void,Void>() {
				Thread worker;
				@Override
				protected Void doInBackground() throws Exception {
					try {
						worker = Thread.currentThread();
						abort_button.addActionListener(new ActionListener() {
							@SuppressWarnings("deprecation")
							public void actionPerformed(ActionEvent e) {
								worker.stop(new TimeoutException());			
							}
						});
						abort_button.setEnabled(true);
						check_button.setEnabled(false);
						parse_button.setEnabled(false);
						analyze_button.setEnabled(false);
						String grammarenc, textenc;
						grammarenc = textenc = Charset.defaultCharset().name();
						String gec = (String)grammarencoding_combobox.getSelectedItem();
						if (!gec.equals("default"))
							grammarenc = gec.substring(0, gec.indexOf(' '));
						String tec = (String)textencoding_combobox.getSelectedItem();
						if (!tec.equals("default"))
							textenc = tec.substring(0, tec.indexOf(' '));
						String grammar, text = null;
						try {
							String grammarloc = grammarfile_textfield.getText();
							if (grammarloc.trim().length() == 0) {
								fail("No grammar specified!");
								return null;
							}
							grammar = Loader.getString(grammarloc, grammarenc);
						} catch (IOException ex) {
							fail("Unable to load grammar: " + ex.getMessage());
							return null;
						}
						if (action.equals("parse")) {
							try {
								String textloc = textfile_textfield.getText();
								if (textloc.trim().length() == 0) {
									fail("No text file specified!");
									return null;
								}
								text = Loader.getString(textloc, textenc);
							} catch (IOException ex) {
								fail("Unable to load text: " + ex.getMessage());
								return null;
							}
						}
						int unfold_level = 0;
						if (action.equals("analyze")) {
							try {
								unfold_level = Integer.parseInt(unfoldlevel_textfield.getText());
							} catch (NumberFormatException e) {
								unfold_level = -1;
							}
						}
						if (unfold_level < 0) {
							fail("Invalid unfold level!");
							return null;
						}
						output_textarea.setText("");
						graphviz_button.setEnabled(false);
						Main.run(grammar, grammarfile_textfield.getText(), 
								text, text != null ? textfile_textfield.getText() : null, action.equals("analyze"), 
								unfold_level, unfoldleft_textfield.getText().trim(), unfoldright_textfield.getText().trim(), 
								verbose_checkbox.isSelected(), dumpast_checkbox.isSelected(), tokenize_checkbox.isSelected(), false, out);
						abort_button.setEnabled(false);
						String output = output_textarea.getText();
						if (output.indexOf("digraph AST {") != -1 && output.endsWith("}" + System.getProperty("line.separator"))) 
							graphviz_button.setEnabled(true);
					} catch (TimeoutException ex) {
						out.println("- execution aborted!");
					} catch (IllegalArgumentException ex) {
						fail(ex.getMessage());
					} catch (InstantiationException ex) {
						fail(ex.getMessage());
					} catch (IllegalAccessException ex) {
						fail(ex.getMessage());
					} catch (ClassNotFoundException ex) {
						fail(ex.getMessage());
					} catch (Exception ex) {
						fail(ex.toString());
					} finally {
						abort_button.setEnabled(false);
						check_button.setEnabled(true);
						parse_button.setEnabled(true);
						analyze_button.setEnabled(true);
					}
					return null;
				}
			}.execute();
		}
		
		private synchronized void graphviz() {
			new SwingWorker<Void,Void>() {
				@Override
				protected Void doInBackground() throws Exception {
					try {
						abort_button.addActionListener(new ActionListener() {
							public void actionPerformed(ActionEvent e) {
								if (dotproc != null) {
									graphviz_aborted = true;
									dotproc.destroy();
								}
							}
						});
						abort_button.setEnabled(true);
						check_button.setEnabled(false);
						parse_button.setEnabled(false);
						analyze_button.setEnabled(false);
						graphviz_button.setEnabled(false);
						String dotcmd = System.getProperty("dk.brics.grammar.dot");
						if (dotcmd == null) {
							fail("Graphviz dot not found!\n" +
							"Use -Ddk.brics.grammar.dot=<path to Graphviz dot> when starting the tool.");
							return null;
						}
						try {
							File dotfile = File.createTempFile("ast", ".dot");
							dotfile.deleteOnExit();
							String dotpath = dotfile.getPath();
							String pngpath = dotpath.substring(0, dotpath.length() - 4) + ".png";
							new File(pngpath).deleteOnExit();
							FileWriter fw = new FileWriter(dotfile);
							String output = output_textarea.getText();
							int dig = output.indexOf("digraph AST {");
							if (dig < 0)
								dig = 0;
							fw.append(output.substring(dig));
							fw.close();
							String[] cmd = { dotcmd, "-Tpng", dotpath, "-o", pngpath};
							output_textarea.setText("executing Graphviz dot... ");
							int status = -1;
							graphviz_aborted = false;
							try {
								dotproc = Runtime.getRuntime().exec(cmd);
								status = dotproc.waitFor();
								dotproc = null;
							} catch (Exception ex) {
								out.println("failed");
								fail("Execution of Graphviz dot failed: " + ex.getMessage());
							}
							if (status == 0) {
								out.println("done");
								out.println("generated image file: " + pngpath);
								if (!ImageViewer.open("AST", pngpath))
									fail("Unable to show image.");
							} else if (graphviz_aborted)
								out.println("aborted");
							else {
								out.println("failed");
								fail("Execution of Graphviz dot failed: return code " + status);
							}
							dotfile.delete();
						} catch (IOException ex) {
							fail("Unable to store temporary file.");
						}
					} finally {
						abort_button.setEnabled(false);
						check_button.setEnabled(true);
						parse_button.setEnabled(true);
						analyze_button.setEnabled(true);
					}
					return null;
				}
			}.execute();
		}

		private void fail(String msg) {
			JOptionPane.showMessageDialog(this, "Error: " + msg, "dk.brics.grammar", JOptionPane.ERROR_MESSAGE);
		}
		

		private void exit() {
			if (JOptionPane.showConfirmDialog(this, "You really want to quit?", 
					"dk.brics.grammar", JOptionPane.YES_NO_OPTION) == JOptionPane.YES_OPTION)
				System.exit(0);
		}
	}

	static final class TextAreaWriter extends Writer {

		private final JTextArea textarea;

		public TextAreaWriter(final JTextArea textArea) {
			this.textarea = textArea;
		}

	    @Override
	    public void flush() {}
	    
	    @Override
	    public void close() {}

		@Override
		public void write(char[] cbuf, int off, int len) {
			textarea.append(new String(cbuf, off, len));
			textarea.setCaretPosition(textarea.getText().length());
		}
	}
}
