package dk.brics.grammar.operations;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.IdentityHashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import dk.brics.automaton.Automaton;

/**
 * Representation of a collection of automata.
 * Singleton automata with equal strings are treated as equal,
 * other automata are compared by object identity.
 * (In other words, automata that are equal but not identical
 * may be represented multiple times.)
 */
public class AutomataCollection implements Iterable<Automaton> {
	
	/**
	 * Empty <code>AutomataCollection</code> (immutable).
	 */
	public static AutomataCollection emptyAutomataCollection = new AutomataCollection(false);
	
	private IdentityHashMap<Automaton,Automaton> h;
	
	private Set<String> singletons;
	
	private boolean modifiable;
	
	/**
	 * Constructs a new empty automata collection.
	 */
	public AutomataCollection() {
		this(true);
	}
	
	private AutomataCollection(boolean modifiable) {
		h = new IdentityHashMap<Automaton,Automaton>();
		singletons = new HashSet<String>();
		this.modifiable = modifiable;
	}
	
	/**
	 * Constructs a new automata collection as a (shallow) copy of an existing one.
	 */
	public AutomataCollection(AutomataCollection ac) {
		h = new IdentityHashMap<Automaton,Automaton>(ac.h);
		singletons = new HashSet<String>(ac.singletons);
		this.modifiable = true;
	}
	
	/**
	 * Adds an automaton.
	 * @param a automaton to add
	 * @return true if this collection changed
	 */
	public boolean add(Automaton a) {
		checkModifiable();
		return addInternal(a);
	}
	
	/**
	 * Adds a collection of automata.
	 * @param as automata to add
	 * @return true if this collection changed
	 */
	public boolean addAll(AutomataCollection as) {
		checkModifiable();
		boolean changed = false;
		if (as != null)
			for (Automaton a : as)
				changed |= addInternal(a);
		return changed;
	}
	
	private final boolean addInternal(Automaton a) {
		String s = a.getSingleton();
		if (s != null) {
			if (singletons.contains(s))
				return false;
			singletons.add(s);
			return true;
		} else
			return h.put(a, a) == null;
	}

	/**
	 * Retains only the automata in this collection that are contained in the specified collection.
	 * @param as automata to be retained (non-null)
	 * @return true if this collection changed
	 */
	public boolean retainAll(AutomataCollection as) {
		checkModifiable();
		boolean changed = singletons.retainAll(as.singletons);
		List<Automaton> remove = new ArrayList<Automaton>();
		for (Automaton a : h.values())
			if (!as.h.containsKey(a))
				remove.add(a);
		if (!remove.isEmpty()) {
			for (Automaton a : remove)
				h.remove(a);
			changed = true;
		}
		return changed;
	}
	
	private final void checkModifiable() {
		if (!modifiable)
			throw new UnsupportedOperationException("automata collection is immutable");
	}

	/**
	 * Returns iterator.
	 * @return iterator
	 */
	public Iterator<Automaton> iterator() {
		return getCollection().iterator();
	}
	
	/**
	 * Returns the collection as a <code>Collection&lt;Automaton&gt;</code>.
	 */
	public Collection<Automaton> getCollection() {
		Collection<Automaton> as = new ArrayList<Automaton>();
		for (String s : singletons)
			as.add(Automaton.makeString(s));
		as.addAll(h.values());
		return as;
	}
	
	/**
	 * Makes this collection immutable.
	 * Subsequent attempts to add automata result in <code>UnsupportedOperationException</code>.
	 */
	public void makeImmutable() {
		modifiable = false;
	}

	/**
	 * Returns a string describing this automata collection.
	 */
	@Override
	public String toString() {
		return getCollection().toString();
	}
}