package dk.brics.grammar.operations;

import dk.brics.automaton.Automaton;
import dk.brics.automaton.RegExp;
import dk.brics.automaton.State;
import dk.brics.automaton.Transition;
import dk.brics.grammar.ambiguity.HorizontalOverlapString;
import dk.brics.grammar.ambiguity.RegularApproximation;

/**
 * Special automata operations used in {@link RegularApproximation}.
 */
public class AutomataOperations {
	
	private static final Automaton marker_sigmaplus = new RegExp("\u0001\u0000(\u0000.)+").toAutomaton();

	private static final Automaton marker_sigmastar = new RegExp("\u0001\u0000(\u0000.)*").toAutomaton();

	private static final Automaton sigmaplus_marker = new RegExp("(\u0000.)+\u0001\u0000").toAutomaton();

	private static final Automaton sigmastar_marker = new RegExp("(\u0000.)*\u0001\u0000").toAutomaton();

	private AutomataOperations() {}
	
	/**
	 * Expands the alphabet of the given alphabet by replacing each transition with two consecutive transitions.
	 * A transition from p to q with label c becomes two transitions (n,c) from p to q via a fresh state. 
	 * @param a input automaton (unmodified)
	 * @param n label for new transitions
	 * @return new automaton
	 */
	public static Automaton expandAlphabet(Automaton a, char n) {
		return expandAlphabet(a, false, n);
	}
		
	private static Automaton expandAlphabet(Automaton a, boolean add_marker, char n) {
		Automaton b;
		if (a.getSingleton() == null || add_marker) {
			b = a.clone();
			for (State p : b.getStates()) {
				State r = new State();
				for (Transition t : p.getTransitions())
					r.addTransition(t);
				p.getTransitions().clear();
				p.addTransition(new Transition(n, r));
				if (add_marker) { // add a marker self loop on each original state
					State q = new State();
					p.addTransition(new Transition('\u0001', q));
					q.addTransition(new Transition('\u0000', p));
				}
			}
		} else {
			StringBuilder t = new StringBuilder();
			for (char c : a.getSingleton().toCharArray())
				t.append('\u0000').append(c);
			b = Automaton.makeString(t.toString());
		}
		return b;
	}
	
	/**
	 * Constructs overlap automaton with expanded alphabet for the two given automata.
	 * All accepted strings contain two (\u005cu0001,\u005cu0000) markers, splitting the string into three parts. 
	 * @param a1 left automaton
	 * @param a2 right automaton
	 * @return overlap automaton 
	 */
	public static Automaton getOverlap(Automaton a1, Automaton a2) {
		Automaton a1f = expandAlphabet(a1, false, '\u0000');
		Automaton a1t = expandAlphabet(a1, true, '\u0000');
		Automaton a2f = expandAlphabet(a2, false, '\u0000');
		Automaton a2t = expandAlphabet(a2, true, '\u0000');
		Automaton b1 = a1f.concatenate(marker_sigmaplus).intersection(a1t).concatenate(marker_sigmastar);
		Automaton b2 = sigmastar_marker.concatenate(sigmaplus_marker.concatenate(a2f).intersection(a2t));
		return b1.intersection(b2);
	}
	
	/**
	 * Finds a shortest example overlap string.
	 * @param overlap overlap automaton
	 * @return overlap string, null if no overlap
	 */
	public static HorizontalOverlapString getOverlapString(Automaton overlap) {
		String s = overlap.getShortestExample(true);
		HorizontalOverlapString h = null;
		StringBuilder b = new StringBuilder();
		if (s != null) {
			int i = 0;
			while (s.charAt(i*2) == '\u0000') {
				b.append(s.charAt(i*2+1));
				i++;
			}
			int i1 = i++;
			while (s.charAt(i*2) == '\u0000') {
				b.append(s.charAt(i*2+1));
				i++;
			}
			int i2 = (i++)-1;
			while (i*2+1 < s.length()) {
				b.append(s.charAt(i*2+1));
				i++;
			}
			h = new HorizontalOverlapString(b.toString(), i1, i2);
		}
		return h;
	}
}
