package dk.brics.grammar.operations;

import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.ListIterator;
import java.util.Set;
import java.util.Stack;

import dk.brics.automaton.Automaton;
import dk.brics.automaton.State;
import dk.brics.automaton.Transition;
import dk.brics.grammar.EOFTerminalEntity;
import dk.brics.grammar.Entity;
import dk.brics.grammar.EntityVisitor;
import dk.brics.grammar.NonterminalEntity;
import dk.brics.grammar.RegexpTerminalEntity;
import dk.brics.grammar.StringTerminalEntity;
import dk.brics.string.Misc;

/**
 * Representation of a set of chars using an ordered list of intervals.
 */
public class CharSet {
	
	private LinkedList<Interval> intervals;
	
	private boolean eof;
	
	/**
	 * Constructs a new empty charset.
	 */
	public CharSet() {
		intervals = new LinkedList<Interval>();
	}
		
	/**
	 * Constructs a charset from the chars occuring in the given string.
	 * @param s string
	 * @param first_only of set, only consider first character
	 * @param last_only of set, only consider last character
	 * @param expanded if set, assume alphabet has been expanded
	 */
	public CharSet(String s, boolean first_only, boolean last_only, boolean expanded) {
		this();
		if (s != null && s.length() > 0)
			if (expanded) {
				if (first_only) {
					int c = two(s.charAt(0), s.charAt(1));
					intervals.add(new Interval(c, c));
				} else if (last_only) {
					int c = two(s.charAt(s.length() - 2), s.charAt(s.length() - 1));
					intervals.add(new Interval(c, c));
				} else {
					for (int i = 0; i < s.length(); i += 2) {
						int c = two(s.charAt(i), s.charAt(i+1));
						intervals.add(new Interval(c, c));
					}
				}
			} else {
				if (first_only) {
					int c = s.charAt(0);
					intervals.add(new Interval(c, c));
				} else if (last_only) {
					int c = s.charAt(s.length() - 1);
					intervals.add(new Interval(c, c));
				} else {
					for (int i = 0; i < s.length(); i++) {
						char c = s.charAt(i);
						intervals.add(new Interval(c, c));
					}
				}
			}
	}
	
	/**
	 * Constructs a charset from the chars occuring in the given automaton.
	 * @param a automaton
	 * @param first_only of set, only consider transitions from the start state
	 * @param last_only of set, only consider transitions to an accept state
	 * @param expanded if set, assume alphabet has been expanded 
	 */
	public CharSet(Automaton a, boolean first_only, boolean last_only, boolean expanded) {
		this(a.getSingleton(), first_only, last_only, expanded);
		if (a.getSingleton() == null) {
			if (expanded) {
				if (first_only) {
					for (Transition t : a.getInitialState().getTransitions())
						for (int c = t.getMin(); c <= t.getMax(); c++)
							for (Transition t2 : t.getDest().getTransitions())
								add(new Interval(two((char)c, t2.getMin()), two((char)c, t2.getMax())));
				} else if (last_only) {
					for (State q : a.getStates())
						for (Transition t : q.getTransitions())
							for (Transition t2 : t.getDest().getTransitions())
								if (t2.getDest().isAccept()) 
									for (int c = t.getMin(); c <= t.getMax(); c++)
										add(new Interval(two((char)c, t2.getMin()), two((char)c, t2.getMax())));
				} else {
					Set<State> visited = new HashSet<State>();
					Stack<State> pending = new Stack<State>(); 
					pending.push(a.getInitialState());
					visited.add(a.getInitialState());
					while (!pending.isEmpty()) {
						State q = pending.pop();
						for (Transition t : q.getTransitions())
							for (int c = t.getMin(); c <= t.getMax(); c++)
								for (Transition t2 : t.getDest().getTransitions()) {
									add(new Interval(two((char)c, t2.getMin()), two((char)c, t2.getMax())));
									if (!visited.contains(t2.getDest())) {
										pending.push(t2.getDest());
										visited.add(t2.getDest());
									}
								}
					}
				}
			} else {
				if (first_only) {
					for (Transition t : a.getInitialState().getTransitions())
						add(new Interval(t.getMin(), t.getMax()));
				} else if (last_only) {
					for (State q : a.getStates())
						for (Transition t : q.getTransitions())
							if (t.getDest().isAccept())
								add(new Interval(t.getMin(), t.getMax()));
				} else {
					for (State q : a.getStates())
						for (Transition t : q.getTransitions())
							add(new Interval(t.getMin(), t.getMax()));
				}
			}
		}
	}

	/**
	 * Adds chars from the given charset to this charset.
	 * @param cs charset
	 * @return true if this charset changed
	 */
	public boolean add(CharSet cs) {
		boolean changed = false;
		for (Interval c : cs.getIntervals())
			changed |= add(c);
		if (cs.containsEOF())
			changed |= addEOF();
		return changed;
	}

	/**
	 * Adds EOF this charset.
	 * @return true if this charset changed
	 */
	public boolean addEOF() {
		boolean changed = !eof;
		eof = true;
		return changed;
	}
	
	/**
	 * Returns the char intervals of this charset.
	 * @return char intervals
	 */
	private LinkedList<Interval> getIntervals() {
		return intervals;
	}
	
	@Override
	public String toString() {
		StringBuilder b = new StringBuilder();
		b.append('[');
		boolean first = true;
		for (Interval i : intervals) {
			if (first)
				first = false;
			else
				b.append(',');
			b.append(i);
		}
		if (eof) {
			if (!first)
				b.append(',');
			b.append("EOF");
		}
		b.append(']');
		return b.toString();
	}
	
	private boolean add(Interval c) {
		boolean changed = false;
		ListIterator<Interval> i = intervals.listIterator();
		ListIterator<Interval> p = intervals.listIterator();
		while (i.hasNext()) {
			Interval ci = i.next();
			if (c.getMax() + 1 < ci.getMin()) {
				p.add(new Interval(c));
				return true;
			}
			if (c.getMin() <= ci.getMax() + 1) {
				if (c.getMin() < ci.getMin()) {
					ci.setMin(c.getMin());
					changed = true;
				}
				if (c.getMax() > ci.getMax()) {
					ci.setMax(c.getMax());
					changed = true;
				}
				p.next();
				while (p.hasNext()) {
					Interval cp = p.next();
					if (ci.getMax() + 1 < cp.getMin())
						return changed;
					if (ci.getMax() >= cp.getMax()) {
						p.remove();
						changed = true;
					} else {
						ci.setMax(cp.getMax());
						p.remove();
						return true;
					}
				}
				return changed;
			}
			p.next();
		}
		i.add(new Interval(c));
		return true;
	}
	
	/**
	 * Checks whether this charset contains the given char.
	 * @param c char
	 * @return true if the char is in the charset
	 */
	public boolean contains(char c) {
		for (Interval ci : intervals)
			if (ci.contains(c))
				return true;
		return false;
	}
	
	/**
	 * Checks whether this charset contains EOF.
	 * @return true if EOF is in the charset
	 */
	public boolean containsEOF() {
		return eof;
	}
	
	/**
	 * Checks whether this charset is disjoint from the given one.
	 * @param cs charset
	 * @return true if disjoint
	 */
	public boolean disjoint(CharSet cs) {
		if (eof && cs.eof)
			return false;
		Iterator<Interval> i1 = intervals.iterator();
		Iterator<Interval> i2 = cs.intervals.iterator();
		if (!i1.hasNext() || !i2.hasNext())
			return true;
		Interval v1 = i1.next();
		Interval v2 = i2.next();
		while (true) {
			if (!v1.disjoint(v2))
				return false;
			if (v1.max < v2.max) {
				if (!i1.hasNext())
					break;
				v1 = i1.next();
			} else {
				if (!i2.hasNext())
					break;
				v2 = i2.next();
			}
		}
		return true;
	}
	
	@Override
	public boolean equals(Object obj) {
		if (obj instanceof CharSet) {
			CharSet c = (CharSet)obj;
			return eof == c.eof && intervals.equals(c.intervals);
		} else
			return false;
	}

	@Override
	public int hashCode() {
		return intervals.hashCode() + (eof ? 1 : 0);
	}

	private int two(char fst, char snd) {
		return fst << 16 ^ snd;
	}

	static class Interval {
		
		private int min, max;

		Interval(int min, int max) {
			this.min = min;
			this.max = max;
		}
		
		Interval(Interval ci) {
			min = ci.min;
			max = ci.max;
		}
		
		boolean contains(int c) {
			return min <= c && c <= max;
		}

		boolean disjoint(Interval ci) {
			return ci.max < min || max < ci.min;
		}

		int getMin() {
			return min;
		}

		int getMax() {
			return max;
		}
		
		void setMin(int min) {
			this.min = min;
		}

		void setMax(int max) {
			this.max = max;
		}
		
		@Override
		public boolean equals(Object obj) {
			if (obj instanceof Interval) {
				Interval c = (Interval)obj;
				return min == c.min && max == c.max;
			} else
				return false;
		}

		@Override
		public int hashCode() {
			return min * 3 + max * 2;
		}

		@Override
		public String toString() {
			if (min == max)
				return Misc.escape(String.valueOf(min));
			else
				return Misc.escape(String.valueOf(min)) + "-" + Misc.escape(String.valueOf(max));
		}
	}

	static CharSet getCharSet(Entity e, final boolean expanded, final boolean first) {
		return e.visitBy(new EntityVisitor<CharSet>() {
			public CharSet visitNonterminalEntity(NonterminalEntity e) {
				return null;
			}
			public CharSet visitRegexpTerminalEntity(RegexpTerminalEntity e) {
				return new CharSet(e.getAutomaton(), first, !first, expanded);
			}
			public CharSet visitStringTerminalEntity(StringTerminalEntity e) {
				if (e.getString().length() == 0)
					return new CharSet();
				else
					return new CharSet(e.getString(), first, !first, expanded);
			}
			public CharSet visitEOFTerminalEntity(EOFTerminalEntity e) {
				CharSet cs = new CharSet();
				cs.addEOF();
				return cs;
			}
		});
	}
}
