package dk.brics.grammar.operations;

import java.util.Arrays;
import java.util.Collection;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import dk.brics.grammar.EOFTerminalEntity;
import dk.brics.grammar.Entity;
import dk.brics.grammar.EntityVisitor;
import dk.brics.grammar.Grammar;
import dk.brics.grammar.GrammarException;
import dk.brics.grammar.NonterminalEntity;
import dk.brics.grammar.Production;
import dk.brics.grammar.RegexpTerminalEntity;
import dk.brics.grammar.StringTerminalEntity;

/**
 * Derives example strings from a grammar.
 */
public class ExampleStringDeriver {
	
	private Grammar g;
	
	private Map<Entity,String> examples;
	
	private Set<Entity> example_contains_eof;
	
	/**
	 * Constructs a new example string deriver.
	 * @param g grammar
	 */
	public ExampleStringDeriver(Grammar g) {
		this.g = g;
		examples = new HashMap<Entity,String>();
		example_contains_eof = new HashSet<Entity>();
	}
	
	/**
	 * Produces an example string that is derivable from the given entity.
	 * @param e entity
	 * @return example string
	 */
	public String getExample(Entity e) {
		return getExample(e, new HashSet<String>());
	}

	private String getExample(final Entity e, final Set<String> active) {
		String example = examples.get(e);
		if (example == null) {
			example = e.visitBy(new EntityVisitor<String>() {
				public String visitNonterminalEntity(NonterminalEntity nt) {
					Production[] ordered = null; 
					String example = nt.getExample();
					if (example == null) {
						Collection<Production> prods = g.getProductions(nt.getNonterminal());
						if (prods == null)
							throw new GrammarException("nonterminal " + nt.getNonterminal() + " has no productions!");
						ordered = prods.toArray(new Production[0]);
						Arrays.sort(ordered, new Comparator<Production>() {
							public int compare(Production p1, Production p2) { // not consistent with equals
								return p2.getPriority() - p1.getPriority();
							}
						});
						// look for a production containing no nonterminals or EOFs
						loop: 
							for (Production p : ordered) {
								StringBuilder ex = new StringBuilder();
								for (Entity e : p.getEntities()) {
									if (e instanceof NonterminalEntity)
										continue loop;
									else {
										ex.append(getExample(e, active));
										if (e instanceof EOFTerminalEntity || example_contains_eof.contains(e)) 
											continue loop;
									}
								}
								example = ex.toString();
								break;
							}
					}
					if (example == null) {
						// look for productive production containing no EOFs
						loop: 
							for (Production p : ordered) {
								StringBuilder ex = new StringBuilder();
								for (Entity e : p.getEntities()) {
									String n = null;
									if (e instanceof NonterminalEntity) { 
										n = ((NonterminalEntity)e).getNonterminal();
										if (active.contains(n))
											continue loop;
										active.add(n);
									}
									ex.append(getExample(e, active));
									if (n != null)
										active.remove(n);
									if (e instanceof EOFTerminalEntity || example_contains_eof.contains(e)) 
										continue loop;					
								}
								example = ex.toString();
								break;
							}
					}
					if (example == null) {
						// look for any productive production
						loop: 
							for (Production p : ordered) {
								boolean contains_eof = false;
								StringBuilder ex = new StringBuilder();
								for (Entity e : p.getEntities()) {
									String n = null;
									if (e instanceof NonterminalEntity) { 
										n = ((NonterminalEntity)e).getNonterminal();
										if (active.contains(n))
											continue loop;
										active.add(n);
									}
									ex.append(getExample(e, active));
									if (n != null)
										active.remove(n);
									if (e instanceof EOFTerminalEntity || example_contains_eof.contains(e))
										contains_eof = true;
								}
								example = ex.toString();
								if (contains_eof)
									example_contains_eof.add(e);
								break;
							}
					}
					if (example == null)
						throw new GrammarException("nonterminal " + nt.getNonterminal() + " is not productive!");
					return example;
				}
				public String visitRegexpTerminalEntity(RegexpTerminalEntity e) {
					String example = e.getExample();
					if (example == null) {
						example = e.getAutomaton().getShortestExample(true);
						if (example == null)
							throw new GrammarException("regular expression has empty language!"); // should have been caught by GrammarChecker
						String ex = example.replace('\t', ' ');
						if (e.getAutomaton().run(ex))
							example = ex;
					}
					return example;
				}
				public String visitStringTerminalEntity(StringTerminalEntity e) {
					return e.getString();
				}
				public String visitEOFTerminalEntity(EOFTerminalEntity e) {
					return "";
				}
			});
			examples.put(e, example);
		}
		return example;
	}
}
