package dk.brics.grammar.operations;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import dk.brics.grammar.Entity;
import dk.brics.grammar.Grammar;
import dk.brics.grammar.NonterminalEntity;
import dk.brics.grammar.Production;
import dk.brics.grammar.TerminalEntity;
import dk.brics.grammar.VoidEntityVisitor;

/**
 * Contains method for finding FIRST and LAST sets.
 * <p>
 * FIRST(N) is the set of chars that may appear as the first char in strings derived from N.
 * LAST(N) is the set of chars that may appear as the last char in strings derived from N.
 * @see SelectFollowsFinder
 */
public class FirstLastFinder {
	
	private Map<Entity,CharSet> entity_first;
	
	private Map<Entity,CharSet> entity_last;
	
	/**
	 * Computes sets for the given grammar.
	 * @param g grammar
	 * @param expanded if set, assume alphabet has been expanded
	 */
	public FirstLastFinder(Grammar g, final boolean expanded) {
		// find possible first/last chars for each terminal entity
		final Map<Entity,CharSet> terminal_first = new HashMap<Entity,CharSet>();
		final Map<Entity,CharSet> terminal_last = new HashMap<Entity,CharSet>();
		for (Production p : g.getProductions())  
			for (Entity e : p.getEntities()) {
				CharSet fi = CharSet.getCharSet(e, expanded, true);
				if (fi != null)
					terminal_first.put(e, fi);
				CharSet la = CharSet.getCharSet(e, expanded, false);
				if (la != null)
					terminal_last.put(e, la);
			}
		// create nodes
		final Map<String,CharSetNode> nonterminal_first = new HashMap<String,CharSetNode>();
		final Map<String,CharSetNode> nonterminal_last = new HashMap<String,CharSetNode>();
		final Map<Production,List<CharSetNode>> production_first = new HashMap<Production,List<CharSetNode>>();
		final Map<Production,List<CharSetNode>> production_last = new HashMap<Production,List<CharSetNode>>();
		Set<CharSetNode> pending = new HashSet<CharSetNode>();
		for (String n : g.getNonterminals()) {
			CharSetNode fi = new CharSetNode();
			pending.add(fi);
			nonterminal_first.put(n, fi);
			CharSetNode la = new CharSetNode();
			pending.add(la);
			nonterminal_last.put(n, la);
		}
		for (Production p : g.getProductions()) { 
			List<CharSetNode> fi = new ArrayList<CharSetNode>();
			List<CharSetNode> la = new ArrayList<CharSetNode>();
			if (p.isUnordered()) { // one shared CharSetNode for all entities (first+last)
				CharSetNode cf = new CharSetNode();
				pending.add(cf);
				fi.add(cf);
				CharSetNode cl = new CharSetNode();
				pending.add(cl);
				la.add(cl);
			} else 
				for (int i = 0; i < p.getEntities().size(); i++) { // one CharSetNode for each entity (first+last)
					CharSetNode cf = new CharSetNode();
					pending.add(cf);
					fi.add(cf);
					CharSetNode cl = new CharSetNode();
					pending.add(cl);
					la.add(cl);
				}
			production_first.put(p, fi);
			production_last.put(p, la);
		}
		// add edges and seeds
		Set<Entity> nullable_entities = new NullableFinder(g).getNullableEntities();
		for (final Production p : g.getProductions()) {
			if (p.getEntities().size() > 0) {
				String n = p.getNonterminal();
				// first_N(a0...an) ==> first(N) if N -> a0...an
				nonterminal_first.get(n).addIn(production_first.get(p).get(0));
				// last_N(a1...an) ==> last(N) if N -> a0...an
				nonterminal_last.get(n).addIn(production_last.get(p).get(production_last.get(p).size() - 1));
			}
			List<Entity> entities = p.getEntities();
			if (p.isUnordered()) {
				// first(ai) ==> first_N(...) if N -> ...ai...
				// last(ai) ==> last_N(...) if N -> ...ai...
				for (int j = 0; j < entities.size(); j++) 
					entities.get(j).visitBy(new VoidEntityVisitor() {
						@Override
						public void visitNonterminal(NonterminalEntity e) {
							production_first.get(p).get(0).addIn(nonterminal_first.get(e.getNonterminal()));
							production_last.get(p).get(0).addIn(nonterminal_last.get(e.getNonterminal()));
						}
						@Override
						public void visitTerminal(TerminalEntity e) {
							production_first.get(p).get(0).getCS().add(terminal_first.get(e));
							production_last.get(p).get(0).getCS().add(terminal_last.get(e));
						}
					});
			} else {
				for (int i = 0; i < entities.size(); i++) {
					final int fi = i;
					// first(aj) ==> first_N(ai...an) if N -> ...ai...an and ai..aj-1 nullable (aj is a terminal or a nonterminal)
					for (int j = i; j < entities.size(); j++) {
						entities.get(j).visitBy(new VoidEntityVisitor() {
							@Override
							public void visitNonterminal(NonterminalEntity e) {
								production_first.get(p).get(fi).addIn(nonterminal_first.get(e.getNonterminal()));
							}
							@Override
							public void visitTerminal(TerminalEntity e) {
								production_first.get(p).get(fi).getCS().add(terminal_first.get(e));
							}
						});
						if (!nullable_entities.contains(entities.get(j)))
							break;
					}
				}
				for (int i = entities.size() - 1; i >= 0; i--) {
					final int fi = i;
					// last(aj) ==> last_N(a1...ai) if N -> a1...ai... and aj+1..ai nullable (aj is a terminal or a nonterminal)
					for (int j = i; j >= 0; j--) {
						entities.get(j).visitBy(new VoidEntityVisitor() {
							@Override
							public void visitNonterminal(NonterminalEntity e) {
								production_last.get(p).get(fi).addIn(nonterminal_last.get(e.getNonterminal()));
							}
							@Override
							public void visitTerminal(TerminalEntity e) {
								production_last.get(p).get(fi).getCS().add(terminal_last.get(e));
							}
						});
						if (!nullable_entities.contains(entities.get(j)))
							break;
					}
				}
			} 
		}
		// compute fixed point
		CharSetNode.fixpoint(pending);
		// extract results
		entity_first = new HashMap<Entity,CharSet>();
		entity_last = new HashMap<Entity,CharSet>();
		for (Production p : g.getProductions()) {
			int i = 0;
			for (Entity e : p.getEntities()) {
				int j = p.isUnordered() ? 0 : i++;
				entity_first.put(e, production_first.get(p).get(j).getCS());
				entity_last.put(e, production_last.get(p).get(j).getCS());
			}
		}
	}
	
	/**
	 * Returns map from entities to charsets representing FIRST information.
	 * @return FIRST sets for entities
	 */
	public Map<Entity,CharSet> getEntityFirst() {
		return entity_first;
	}
	
	/**
	 * Returns map from entities to charsets representing LAST information.
	 * @return LAST sets for entities
	 */
	public Map<Entity,CharSet> getEntityLast() {
		return entity_last;
	}
}

