package dk.brics.grammar.operations;

import java.util.HashMap;
import java.util.List;
import java.util.Map;

import dk.brics.automaton.Automaton;
import dk.brics.grammar.EOFTerminalEntity;
import dk.brics.grammar.Entity;
import dk.brics.grammar.EntityVisitor;
import dk.brics.grammar.Grammar;
import dk.brics.grammar.Production;
import dk.brics.grammar.NonterminalEntity;
import dk.brics.grammar.RegexpTerminalEntity;
import dk.brics.grammar.StringTerminalEntity;
import dk.brics.string.grammar.Nonterminal;

/**
 * Converter from normalized grammars to JSA grammars.
 * @see ProductionLengthNormalizer
 */
public class Grammar2JSAGrammar {
	
	private Map<String,dk.brics.string.grammar.Nonterminal> nt_map;
	
	/** 
	 * Constructs a new converter.
	 */
	public Grammar2JSAGrammar() {}
	
	/**
	 * Converts a normalized grammar to JSA grammar representation.
	 * @param ng grammar to convert
	 * @return JSA grammar
	 * @throws IllegalArgumentException if the input grammar is not normalized
	 */
	public dk.brics.string.grammar.Grammar convert(Grammar ng) throws IllegalArgumentException {
		dk.brics.string.grammar.Grammar jg = new dk.brics.string.grammar.Grammar();
		nt_map = new HashMap<String,dk.brics.string.grammar.Nonterminal>();
		for (Production p : ng.getProductions()) {
			List<Entity> entities = p.getEntities();
			dk.brics.string.grammar.Nonterminal jn = getNonterminal(p.getNonterminal(), jg);
			if (entities.size() == 0)
				jg.addEpsilonProduction(jn);
			else if (entities.size() == 1)
				jg.addUnitProduction(jn, convertEntity(entities.get(0), jg));
			else if (entities.size() == 2)
				jg.addPairProduction(jn, convertEntity(entities.get(0), jg), convertEntity(entities.get(1), jg));
			else
				throw new IllegalArgumentException("grammar is not normalized");
		}
		return jg;
	}
	
	/**
	 * Looks up or creates nonterminal.
	 */
	private dk.brics.string.grammar.Nonterminal getNonterminal(String n, dk.brics.string.grammar.Grammar jg) {
		dk.brics.string.grammar.Nonterminal jn = nt_map.get(n);
		if (jn == null)
			jn = jg.addNonterminal();
		nt_map.put(n, jn);
		return jn;
	}
	
	/**
	 * Converts entity to nonterminal.
	 */
	private dk.brics.string.grammar.Nonterminal convertEntity(Entity e, final dk.brics.string.grammar.Grammar jg) {
		return e.visitBy(new EntityVisitor<dk.brics.string.grammar.Nonterminal>() {
			public Nonterminal visitNonterminalEntity(NonterminalEntity e) {
				return getNonterminal(e.getNonterminal(), jg);
			}
			public Nonterminal visitRegexpTerminalEntity(RegexpTerminalEntity e) {
				dk.brics.string.grammar.Nonterminal jn = jg.addNonterminal();
				jg.addAutomatonProduction(jn, e.getAutomaton());
				return jn;
			}
			public Nonterminal visitStringTerminalEntity(StringTerminalEntity e) {
				dk.brics.string.grammar.Nonterminal jn = jg.addNonterminal();
				jg.addAutomatonProduction(jn, Automaton.makeString(e.getString()));
				return jn;
			}
			public Nonterminal visitEOFTerminalEntity(EOFTerminalEntity e) {
				dk.brics.string.grammar.Nonterminal jn = jg.addNonterminal();
				jg.addEpsilonProduction(jn); // sound to treat EOF as ""
				return jn;
			}
		});
	}

	/**
	 * Returns map from old nonterminals to new nonterminals from the last conversion.
	 * @return nonterminal map
	 */
	public Map<String,dk.brics.string.grammar.Nonterminal> getNonterminalMap() {
		return nt_map;
	}
}
