package dk.brics.grammar.operations;

import java.io.PrintWriter;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.Map;
import java.util.Set;

import dk.brics.grammar.EOFTerminalEntity;
import dk.brics.grammar.Entity;
import dk.brics.grammar.EntityVisitor;
import dk.brics.grammar.Grammar;
import dk.brics.grammar.NonterminalEntity;
import dk.brics.grammar.Production;
import dk.brics.grammar.RegexpTerminalEntity;
import dk.brics.grammar.StringTerminalEntity;
import dk.brics.grammar.VoidEntityVisitor;

/**
 * Checks grammars for nonproductive or unreachable nonterminals.
 */
public class GrammarChecker {

	private Grammar g;
	
	/** 
	 * Constructs a new grammar checker.
	 */
	public GrammarChecker() {}
	
	/**
	 * Checks the given grammar.
	 * @param g grammar
	 * @param out warnings are printed to this stream
	 * @return number of warnings produced
	 */
	public int check(final Grammar g, PrintWriter out) {
		this.g = g;
		int warnings = 0;
		// find nonterminals with no productions
		final Set<String> noprods = new HashSet<String>();
		for (Production p : g.getProductions())
			for (Entity e : p.getEntities()) 
				e.visitBy(new VoidEntityVisitor() {
					@Override
					public void visitNonterminal(NonterminalEntity e) {
						if (g.getProductions(e.getNonterminal()) == null)
							noprods.add(e.getNonterminal());
					}
				});
		for (String n : noprods) {
			out.println("*** nonterminal '" + n + "' has no productions");
			warnings++;
		}
		// find unreachable nonterminals
		Set<String> reachable = new HashSet<String>();
		findReachable(g.getStart(), reachable);
		Set<String> unreachable = new HashSet<String>(g.getNonterminals());
		unreachable.removeAll(reachable);
		for (String n : unreachable) {
			out.println("*** nonterminal '" + n + "' is not reachable from the start nonterminal");
			warnings++;
		}
		// find nonterminals that have productions but are non-productive
		Set<String> productive = new HashSet<String>();
		for (String n : reachable)
			if (!noprods.contains(n) && !checkProductive(n, productive, new HashSet<String>())) {
				out.println("*** nonterminal '" + n + "' cannot derive any strings");
				warnings++;
			}
		// find nonterminals with duplicate production labels
		for (String n : g.getNonterminals()) {
			Set<String> labels = new HashSet<String>();
			for (Production p : g.getProductions(n)) {
				String l = p.getID().getLabel();
				if (l != null)
					if (labels.contains(l)) {
						out.println("*** multiple productions for '" + p.getNonterminal() + "' with label '" + l + "'");
						warnings++;				
					} else
						labels.add(l);
			}
		}
		// find nonterminals that are non-productive because they contain a must-EOF entity followed by a non-nullable entity (and are ordered)
		Map<String,Set<Production>> reverse = buildReverseEdgess();
		Set<String> musteof = getMustEOFNonterminals(reverse);
		Set<Entity> nullable = new NullableFinder(g).getNullableEntities();
		Set<String> bad = new HashSet<String>();
		for (Production p : g.getProductions())
			if (!p.isUnordered()) {
				boolean seen_eof = false;
				for (Entity e : p.getEntities())
					if (!seen_eof && 
							(e instanceof EOFTerminalEntity || 
									(e instanceof NonterminalEntity && musteof.contains(((NonterminalEntity)e).getNonterminal()))))
						seen_eof = true;
					else if (seen_eof && !nullable.contains(e))
						bad.add(p.getNonterminal());
			}
		for (String n : bad) {
			out.println("*** nonterminal '" + n + "' cannot derive any strings (due to EOF terminals)");
			warnings++;
		}
		return warnings;
	}
	
	private void findReachable(String n, final Set<String> reachable) {
		if (!reachable.contains(n)) {
			reachable.add(n);
			if (g.getProductions(n) != null)
				for (Production p : g.getProductions(n))
					for (Entity e : p.getEntities()) 
						e.visitBy(new VoidEntityVisitor() {
							@Override
							public void visitNonterminal(NonterminalEntity e) {
								findReachable(e.getNonterminal(), reachable);
							}
						});
		}
	}
	
	private boolean checkProductive(String n, final Set<String> productive, final Set<String> visited) {
		boolean prod;
		if (visited.contains(n)) 
			prod = productive.contains(n);
		else {	
			visited.add(n);
			prod = false;
			if (g.getProductions(n) != null)
				for (Production p : g.getProductions(n)) {
					boolean allprod = true;
					for (Entity e : p.getEntities())
						allprod &= e.visitBy(new EntityVisitor<Boolean>() {
							public Boolean visitNonterminalEntity(NonterminalEntity e) {
								return checkProductive(e.getNonterminal(), productive, visited);
							}
							public Boolean visitRegexpTerminalEntity(RegexpTerminalEntity e) {
								return !e.getAutomaton().isEmpty();
							}
							public Boolean visitStringTerminalEntity(StringTerminalEntity e) {
								return true;
							}
							public Boolean visitEOFTerminalEntity(EOFTerminalEntity e) {
								return true;
							}
						});
					if (allprod) {
						productive.add(n);
						prod = true;
						break;
					}
				}
		}
		return prod;
	}
	
	/**
	 * Builds map from nonterminal to productions containing that nonterminal on the right-hand side.
	 */
	private Map<String,Set<Production>> buildReverseEdgess() {
		final Map<String,Set<Production>> reverse = new HashMap<String,Set<Production>>();
		for (final Production p : g.getProductions())
			for (Entity e : p.getEntities())
				e.visitBy(new VoidEntityVisitor() {
					@Override
					public void visitNonterminal(NonterminalEntity e) {
						Set<Production> prods = reverse.get(e.getNonterminal());
						if (prods == null) {
							prods = new HashSet<Production>();
							reverse.put(e.getNonterminal(), prods);
						}
						prods.add(p);
					}
				});
		return reverse;
	}
	
	/**
	 * Finds nonterminals where every derivable string contains EOF.
	 */
	private Set<String> getMustEOFNonterminals(Map<String,Set<Production>> reverse) {
		Set<String> musteof = new HashSet<String>();
		LinkedHashSet<String> pending = new LinkedHashSet<String>(g.getNonterminals());
		while (!pending.isEmpty()) {
			String nt = pending.iterator().next();
			pending.remove(nt);
			boolean allprods = true, anyprods = false;
			for (Production p : g.getProductions(nt)) {
				boolean anyents = false;
				for (Entity e : p.getEntities())
					if (e instanceof EOFTerminalEntity || 
							(e instanceof NonterminalEntity && musteof.contains(((NonterminalEntity)e).getNonterminal())))
						anyents = true;
				if (anyents) 
					anyprods = true;
				else
					allprods = false;
			}
			if (allprods && anyprods) 
				if (musteof.add(nt)) {
					Set<Production> prevs = reverse.get(nt);
					if (prevs != null)
						for (Production p : prevs)
							pending.add(p.getNonterminal());
				}
		}
		return musteof;
	}
}
