package dk.brics.grammar.operations;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.Stack;

import dk.brics.grammar.EOFTerminalEntity;
import dk.brics.grammar.Entity;
import dk.brics.grammar.EntityVisitor;
import dk.brics.grammar.Grammar;
import dk.brics.grammar.NonterminalEntity;
import dk.brics.grammar.Production;
import dk.brics.grammar.RegexpTerminalEntity;
import dk.brics.grammar.StringTerminalEntity;
import dk.brics.grammar.VoidEntityVisitor;

/**
 * Finds entities, nonterminals, and productions that can derive the empty string.
 */
public class NullableFinder {
	
	private Set<Entity> nullable_entities;
	
	private Set<String> nullable_nonterminals;
	
	private Set<Production> nullable_productions;
	
	/** 
	 * Finds the nullables for the given grammar.
	 */
	public NullableFinder(Grammar g) { 
		nullable_entities = new HashSet<Entity>();
		nullable_nonterminals = new HashSet<String>();
		nullable_productions = new HashSet<Production>();
		Stack<String> pending = new Stack<String>();
		final Map<String,Set<Production>> nt2prod = new HashMap<String,Set<Production>>();
		// find nullable terminal entities, seed pending, and find back edges
		for (final Production p : g.getProductions()) {
			boolean all_nullable_termininals = true;
			for (Entity e : p.getEntities())
				if(e.visitBy(new EntityVisitor<Boolean>() {
					public Boolean visitNonterminalEntity(NonterminalEntity e) {
						Set<Production> ps = nt2prod.get(e.getNonterminal());
						if (ps == null) {
							ps = new HashSet<Production>();
							nt2prod.put(e.getNonterminal(), ps);
						}
						ps.add(p);
						return false;
					}
					public Boolean visitRegexpTerminalEntity(RegexpTerminalEntity e) {
						return e.getAutomaton().run("");
					}
					public Boolean visitStringTerminalEntity(StringTerminalEntity e) {
						return e.getString().length() == 0;
					}
					public Boolean visitEOFTerminalEntity(EOFTerminalEntity e) {
						return true;
					}
				})) 
					nullable_entities.add(e);
				else
					all_nullable_termininals = false;
			if (all_nullable_termininals)
				pending.push(p.getNonterminal());
		}
		// find nullable nonterminals by iterating until fixpoint
		while (!pending.isEmpty()) {
			String n = pending.pop();
			nullable_nonterminals.add(n);
			if (nt2prod.get(n) != null)
				for (Production p : nt2prod.get(n)) {
					boolean all = true;
					for (Entity e : p.getEntities()) 
						all &= e.visitBy(new EntityVisitor<Boolean>() {
							public Boolean visitNonterminalEntity(NonterminalEntity e) {
								return nullable_nonterminals.contains(e.getNonterminal());
							}
							public Boolean visitRegexpTerminalEntity(RegexpTerminalEntity e) {
								return nullable_entities.contains(e);
							}
							public Boolean visitStringTerminalEntity(StringTerminalEntity e) {
								return nullable_entities.contains(e);
							}
							public Boolean visitEOFTerminalEntity(EOFTerminalEntity e) {
								return nullable_entities.contains(e);
							}
						});
					if (all && !nullable_nonterminals.contains(p.getNonterminal()))
						pending.push(p.getNonterminal());
				}
		}
		// collect nullable nonterminal entities
		for (Production p : g.getProductions())
			for (Entity e : p.getEntities())
				e.visitBy(new VoidEntityVisitor() {
					@Override public void visitNonterminal(NonterminalEntity e) {
						if (nullable_nonterminals.contains(e.getNonterminal()))
								nullable_entities.add(e);
					}
				});
		// collect nullable productions
		for (Production p : g.getProductions()) 
			if (nullable_entities.containsAll(p.getEntities()))
				nullable_productions.add(p);
	}

	/**
	 * Returns the set of entities that can derive the empty string.
	 * @return set of entities
	 */
	public Set<Entity> getNullableEntities() {
		return nullable_entities;
	}

	/**
	 * Returns the set of nonterminals that can derive the empty string.
	 * @return set of nonterminals
	 */
	public Set<String> getNullableNonterminals() {
		return nullable_nonterminals;
	}

	/**
	 * Returns the set of productions that can derive the empty string.
	 * @return set of productions
	 */
	public Set<Production> getNullableProductions() {
		return nullable_productions;
	}
}
