package dk.brics.grammar.operations;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import dk.brics.grammar.Entity;
import dk.brics.grammar.Grammar;
import dk.brics.grammar.NonterminalEntity;
import dk.brics.grammar.Production;
import dk.brics.grammar.TerminalEntity;
import dk.brics.grammar.VoidEntityVisitor;

/**
 * Contains method for finding SELECT and FOLLOWS sets.
 * <p>
 * FOLLOWS(N) is the set of chars that can appear immediately after a string derived from N in
 * derivations from the start nonterminal.
 * SELECT(N) is here the set of chars that may appear as the first char in strings derived from N,
 * plus - if N is nullable - the chars in FOLLOWS(N).
 * @see FirstLastFinder
 */
public class SelectFollowsFinder {
	
	private Map<String,CharSet> select;
	
	private Map<String,CharSet> follows;
	
	/**
	 * Computes sets for the given grammar.
	 * @param g grammar
	 */
	public SelectFollowsFinder(Grammar g) {
		// find possible first chars for each terminal entity
		final Map<Entity,CharSet> terminal_first = new HashMap<Entity,CharSet>();
		for (Production p : g.getProductions())  
			for (Entity e : p.getEntities()) {
				CharSet fi = CharSet.getCharSet(e, false, true);
				if (fi != null)
					terminal_first.put(e, fi);
			}
		// create nodes
		final Map<String,CharSetNode> nonterminal_follows = new HashMap<String,CharSetNode>();
		final Map<String,CharSetNode> nonterminal_select = new HashMap<String,CharSetNode>();
		final Map<Production,List<CharSetNode>> production_select = new HashMap<Production,List<CharSetNode>>();
		Set<CharSetNode> pending = new HashSet<CharSetNode>();
		for (String n : g.getNonterminals()) {
			CharSetNode fo = new CharSetNode();
			pending.add(fo);
			nonterminal_follows.put(n, fo);
			CharSetNode fi = new CharSetNode();
			pending.add(fi);
			nonterminal_select.put(n, fi);
		}
		for (Production p : g.getProductions()) { 
			List<CharSetNode> ln = new ArrayList<CharSetNode>();
			if (p.isUnordered()) { // one shared CharSetNode for all production points
				CharSetNode f = new CharSetNode();
				pending.add(f);
				ln.add(f);
			} else 
				for (int i = 0; i <= p.getEntities().size(); i++) { // one CharSetNode for each production point
					CharSetNode f = new CharSetNode();
					pending.add(f);
					ln.add(f);
				}
			production_select.put(p, ln);
		}
		// add edges and seeds
		Set<Entity> nullable_entities = new NullableFinder(g).getNullableEntities();
		for (final Production p : g.getProductions()) {
			String n = p.getNonterminal();
			// select_N(a1...an) ==> select(N) if N -> a1...an
			nonterminal_select.get(n).addIn(production_select.get(p).get(0));
			List<Entity> entities = p.getEntities();
			if (p.isUnordered()) {
				// select_N(...) ==> follows(M) if N -> ...M... 
				for (int i = 0; i < entities.size(); i++) 
					entities.get(i).visitBy(new VoidEntityVisitor() {
						@Override
						public void visitNonterminal(NonterminalEntity e) {
							if (nonterminal_follows.get(e.getNonterminal()) != null)
								nonterminal_follows.get(e.getNonterminal()).addIn(production_select.get(p).get(0));
						}
					});
				// follows(N) ==> select_N(...)
				production_select.get(p).get(0).addIn(nonterminal_follows.get(n));
				// select(ai) ==> select_N(...) if N -> ...ai...
				for (int j = 0; j < entities.size(); j++) 
					entities.get(j).visitBy(new VoidEntityVisitor() {
						@Override
						public void visitNonterminal(NonterminalEntity e) {
							production_select.get(p).get(0).addIn(nonterminal_select.get(e.getNonterminal()));
						}
						@Override
						public void visitTerminal(TerminalEntity e) {
							production_select.get(p).get(0).getCS().add(terminal_first.get(e));
						}
					});
			} else {
				// follows(N) ==> select_N(epsilon)
				production_select.get(p).get(entities.size()).addIn(nonterminal_follows.get(n));
				for (int i = 0; i < entities.size(); i++) {
					final int fi = i;
					entities.get(i).visitBy(new VoidEntityVisitor() {
						@Override
						public void visitNonterminal(NonterminalEntity e) {
							// select_N(ai...an) ==> follows(M) if N -> ...Mai...an 
							if (nonterminal_follows.get(e.getNonterminal()) != null)
								nonterminal_follows.get(e.getNonterminal()).addIn(production_select.get(p).get(fi + 1));
						}
					});
					// follows(N) ==> select_N(ai..an) if N -> ...ai...an and ai...an nullable
					boolean nullable = true;
					for (int j = i; j < entities.size(); j++) 
						if (!nullable_entities.contains(entities.get(j))) {
							nullable = false;
							break;
						}
					if (nullable) 
						production_select.get(p).get(i).addIn(nonterminal_follows.get(n));
					// select(aj) ==> select_N(ai...an) if N -> ...ai...an and ai..aj-1 nullable (aj is a terminal or a nonterminal)
					for (int j = i; j < entities.size(); j++) {
						entities.get(j).visitBy(new VoidEntityVisitor() {
							@Override
							public void visitNonterminal(NonterminalEntity e) {
								production_select.get(p).get(fi).addIn(nonterminal_select.get(e.getNonterminal()));
							}
							@Override
							public void visitTerminal(TerminalEntity e) {
								production_select.get(p).get(fi).getCS().add(terminal_first.get(e));							}
						});
						if (!nullable_entities.contains(entities.get(j)))
							break;
					}
				}
			} 
		}
		// EOF ==> follows(S)
		nonterminal_follows.get(g.getStart()).getCS().addEOF();
		// compute fixed point
		CharSetNode.fixpoint(pending);
		// extract results
		select = new HashMap<String,CharSet>();
		follows = new HashMap<String,CharSet>();
		for (String n : g.getNonterminals()) {
			select.put(n, nonterminal_select.get(n).getCS());
			follows.put(n, nonterminal_follows.get(n).getCS());
		}		
	}
	
	/**
	 * Returns map from nonterminal names to charsets representing SELECT information.
	 * @return SELECT sets
	 */
	public Map<String,CharSet> getNonterminalSelect() {
		return select;
	}
	
	/**
	 * Returns map from nonterminal names to charsets representing FOLLOWS information.
	 * @return FOLLOWS sets
	 */
	public Map<String,CharSet> getNonterminalFollows() {
		return follows;
	}
}
