package dk.brics.grammar.operations;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.Stack;

import dk.brics.automaton.Automaton;
import dk.brics.grammar.EOFTerminalEntity;
import dk.brics.grammar.Entity;
import dk.brics.grammar.EntityVisitor;
import dk.brics.grammar.Grammar;
import dk.brics.grammar.NonterminalEntity;
import dk.brics.grammar.Production;
import dk.brics.grammar.RegexpTerminalEntity;
import dk.brics.grammar.StringTerminalEntity;
import dk.brics.grammar.TerminalEntity;
import dk.brics.grammar.VoidEntityVisitor;

/**
 * Finds terminals that may/must occur in derivable strings.
 * <p>
 * A "may" terminal is one that might occur in some derivation (so the set of derivable strings
 * is included in the Kleene star of the union of the may-terminals).
 * A "must" terminal is one that definitely occurs in every derivation of a string (so for every must-terminal,
 * every derivable string contains a substring that matches the must-terminal). 
 * EOF terminals are excluded in both.
 */
public class TerminalFinder {

	private Map<String,AutomataCollection> nonterminal_may;

	private Map<String,AutomataCollection> nonterminal_must;

	private Map<Production,AutomataCollection> production_may;

	private Map<Production,AutomataCollection> production_must;
	
	/**
	 * Finds the may/must terminals for the given grammar.
	 */
	public TerminalFinder(Grammar g) {
		nonterminal_may = new HashMap<String,AutomataCollection>();
		nonterminal_must = new HashMap<String,AutomataCollection>();
		production_may = new HashMap<Production,AutomataCollection>();
		production_must = new HashMap<Production,AutomataCollection>();
		final Map<TerminalEntity,Automaton> t2a = new HashMap<TerminalEntity,Automaton>();
		final Map<String,Set<Production>> nt2prod = new HashMap<String,Set<Production>>();
		for (String n : g.getNonterminals()) 
			nonterminal_may.put(n, new AutomataCollection());
		Set<String> pendingset = new HashSet<String>();
		Stack<String> pendinglist = new Stack<String>();
		// find terminals, seed pending and find back edges
		for (final Production p : g.getProductions()) {
			String pn = p.getNonterminal();
			AutomataCollection p_may = new AutomataCollection();
			production_may.put(p, p_may);
			AutomataCollection pn_may = nonterminal_may.get(pn);
			for (Entity e : p.getEntities()) {
				Automaton a = e.visitBy(new EntityVisitor<Automaton>() {
					public Automaton visitNonterminalEntity(NonterminalEntity e) {
						Set<Production> ps = nt2prod.get(e.getNonterminal());
						if (ps == null) {
							ps = new HashSet<Production>();
							nt2prod.put(e.getNonterminal(), ps);
						}
						ps.add(p);
						return null;
					}
					public Automaton visitRegexpTerminalEntity(RegexpTerminalEntity e) {
						return e.getAutomaton();
					}
					public Automaton visitStringTerminalEntity(StringTerminalEntity e) {
						return Automaton.makeString(e.getString());
					}
					public Automaton visitEOFTerminalEntity(EOFTerminalEntity e) {
						return null; // ignoring EOF terminals
					}}
				);
				if (a != null) {
					t2a.put((TerminalEntity)e, a);
					pn_may.add(a);
					p_may.add(a);
				}
			}
		}
		// find may productions/nonterminals by iterating until fixpoint
		pendingset.addAll(g.getNonterminals());
		pendinglist.addAll(g.getNonterminals());
		while (!pendinglist.isEmpty()) {
			String n = pendinglist.pop();
			pendingset.remove(n);
			AutomataCollection n_may = nonterminal_may.get(n);
			if (nt2prod.get(n) != null)
				for (Production p : nt2prod.get(n)) {
					String pn = p.getNonterminal();
					production_may.get(p).addAll(n_may);
					boolean pn_changed = nonterminal_may.get(pn).addAll(n_may);
					if (pn_changed && !pendingset.contains(pn)) {
						pendingset.add(pn);
						pendinglist.add(pn);
					}
				}					
		}
		// find must productions/nonterminals by iterating coinductively until fixpoint
		for (String n : g.getNonterminals()) 
			nonterminal_must.put(n, new AutomataCollection(nonterminal_may.get(n)));
		for (Production p : g.getProductions())
			production_must.put(p, new AutomataCollection(production_may.get(p)));
		pendingset.addAll(g.getNonterminals());
		pendinglist.addAll(g.getNonterminals());
		while (!pendinglist.isEmpty()) {
			String n = pendinglist.pop();
			pendingset.remove(n);
			AutomataCollection n_must = nonterminal_must.get(n);
			boolean n_changed = false;
			for (Production p : g.getProductions(n))
				n_changed |= n_must.retainAll(production_must.get(p));
			if (n_changed && nt2prod.get(n) != null)
				for (Production p : nt2prod.get(n)) {
					String pn = p.getNonterminal();
					AutomataCollection p_must = production_must.get(p);
					final AutomataCollection new_p_must = new AutomataCollection();
					for (Entity e : p.getEntities()) 
						e.visitBy(new VoidEntityVisitor() {
							@Override
							public void visitNonterminal(NonterminalEntity e) {
								new_p_must.addAll(nonterminal_must.get(e.getNonterminal())); 
							}
							@Override
							public void visitTerminal(TerminalEntity e) {
								new_p_must.add(t2a.get(e));
							}
						});
					boolean pn_changed = p_must.retainAll(new_p_must);
					if (pn_changed && !pendingset.contains(pn)) {
						pendingset.add(pn);
						pendinglist.add(pn);
					}
				}
		}
	}

	/**
	 * Returns the terminals that appear in some derivations from the given nonterminal. 
	 */
	public AutomataCollection getMayTerminals(String nt) {
		AutomataCollection c = nonterminal_may.get(nt);
		if (c == null)
			return AutomataCollection.emptyAutomataCollection;
		return c;
	}

	/**
	 * Returns the terminals that appear in all derivations from the given nonterminal. 
	 */
	public AutomataCollection getMustTerminals(String nt) {
		AutomataCollection c = nonterminal_must.get(nt);
		if (c == null)
			return AutomataCollection.emptyAutomataCollection;
		return c;
	}

	/**
	 * Returns the terminals that appear in some derivations from the given production. 
	 */
	public AutomataCollection getMayTerminals(Production p) {
		AutomataCollection c = production_may.get(p);
		if (c == null)
			return AutomataCollection.emptyAutomataCollection;
		return c;
	}

	/**
	 * Returns the terminals that appear in all derivations from the given production. 
	 */
	public AutomataCollection getMustTerminals(Production p) {
		AutomataCollection c = production_must.get(p);
		if (c == null)
			return AutomataCollection.emptyAutomataCollection;
		return c;
	}
}
