package dk.brics.grammar.operations;

import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.List;
import java.util.Stack;

import dk.brics.automaton.Automaton;
import dk.brics.automaton.State;
import dk.brics.automaton.Transition;
import dk.brics.grammar.EOFTerminalEntity;
import dk.brics.grammar.Entity;
import dk.brics.grammar.EntityVisitor;
import dk.brics.grammar.Grammar;
import dk.brics.grammar.NonterminalEntity;
import dk.brics.grammar.Production;
import dk.brics.grammar.RegexpTerminalEntity;
import dk.brics.grammar.StringTerminalEntity;

// TODO: support multi-character parentheses? parentheses in regexp terminals?
// TODO: auto-detect unfolding command

// TODO: if not balanced, then try to convert it to being balanced (using Knuth's algorithms?)

/**
 * Grammar unfolder.
 */
public class Unfolder {
	
	/** 
	 * Print stream for messages. 
	 */
	private PrintWriter out;
	
	/**
	 * Constructs a new unfolder.
	 * @param out print writer for messages
	 */
	public Unfolder(PrintWriter out) {
		this.out = out;
	}
	
	/**
	 * Unfolds the given grammar.
	 * @param g grammar
	 * @param number number of times to unfold
	 * @param lpar left parenthesis symbols
	 * @param rpar right parenthesis symbols (in same order as the corresponding left parentheses), disjoint from left parenthesis symbols
	 * @return new grammar
	 * @throws IllegalArgumentException if the grammar cannot be unfolded with the given parentheses
	 */
	public Grammar unfold(Grammar g, int number, final String lpar, final String rpar) throws IllegalArgumentException {
		List<Production> allproductions = new ArrayList<Production>();
		boolean any = false;
		// create number+1 copies of the grammar, expand terminal alphabet and nonterminal names,
		// and shift level at entities "between parentheses"
		final Stack<Character> parstack = new Stack<Character>();
		for (int i = 0; i <= number; i++) {
			List<Production> thislevelprods = new ArrayList<Production>();
			for (final Production p : g.getProductions()) {
				List<Entity> entities = p.getEntities();
				// build new entities
				final List<Entity> newentities = new ArrayList<Entity>();
				for (Entity e : entities) {
					int step = e.visitBy(new EntityVisitor<Integer>() {
						public Integer visitNonterminalEntity(NonterminalEntity e) {
							return 0;
						}
						public Integer visitRegexpTerminalEntity(RegexpTerminalEntity e) {
							if (contains(e.getAutomaton(), rpar))
								throw new IllegalArgumentException(productionName(p) + ": can't unfold grammar with parenthesis in regexp terminal");
							return 0;
						}
						public Integer visitStringTerminalEntity(StringTerminalEntity e) {
							String s = e.getString();
							if (s.length() == 1) {
								char c = s.charAt(0);
								if (contains(s, lpar)) {
									parstack.push(c);
									return 1;
								} else if (contains(s, rpar)) {
									if (parstack.isEmpty() || parstack.peek() != lpar.charAt(rpar.indexOf(c)))
										throw new IllegalArgumentException(productionName(p) + ": can't unfold grammar with non-balanced parentheses");
									parstack.pop();
									return 0;
								}
							} else if (contains(s, rpar) || contains(s, lpar)) {
								if (s.length() == 2) {
									char c1 = s.charAt(0);
									char c2 = s.charAt(1);
									int i1 = lpar.indexOf(c1);
									int i2 = rpar.indexOf(c2);
									if (i1 >=0 && i1 == i2)
										return 0;
								}
								throw new IllegalArgumentException(productionName(p) + ": can't unfold grammar with parenthesis in non-singleton string terminal");
							}
							return 0;
						}
						public Integer visitEOFTerminalEntity(EOFTerminalEntity e) {
							return 0;
						}
					});
					if (!parstack.isEmpty()) {
						if (p.isUnordered())
							throw new IllegalArgumentException(productionName(p) + ": can't unfold grammar with unordered parenthesis productions");
						any = true;
					}
					final int level = (parstack.size() - step + i < number) ? parstack.size() - step + i : number;
					newentities.add(e.visitBy(new EntityVisitor<Entity>() {
						public Entity visitNonterminalEntity(NonterminalEntity e) {
							return new NonterminalEntity(expandName(e.getNonterminal(), level), e.getLabel(), e.getExample());
						}
						public Entity visitRegexpTerminalEntity(RegexpTerminalEntity e) {
							return new RegexpTerminalEntity(AutomataOperations.expandAlphabet(e.getAutomaton(), (char)level), 
									e.isMax(), e.getAutomatonName(), e.getLabel(), null); // note: removing example strings
						}
						public Entity visitStringTerminalEntity(StringTerminalEntity e) {
							return new StringTerminalEntity(expandAlphabet(e.getString(), (char)level));
						}
						public Entity visitEOFTerminalEntity(EOFTerminalEntity e) {
							return new EOFTerminalEntity();
						}
					}));
				}
				if (!parstack.isEmpty())
					throw new IllegalArgumentException(productionName(p) + ": can't unfold grammar with non-balanced parentheses");
				Production np = new Production(expandName(p.getNonterminal(), i), newentities, p.isUnordered(), p.getID(), p.getPriority());
				if (i > 0)
					np.setUnfolded();
				allproductions.add(np);
				thislevelprods.add(np);
			}
		}
		if (!any)
			out.println("*** warning: no productions matching the given unfolding parentheses");
		Grammar u = new Grammar(expandName(g.getStart(), 0), allproductions);
		u.setUnfolded(true);
		return u;
	}
	
	private String productionName(Production p) {
		return "production " + p.getNonterminal() + "[" + p.getID().getLabel() + "]";
	}

	private String expandAlphabet(String s, char level) {
		StringBuilder b = new StringBuilder();
		for (int i = 0; i < s.length(); i++)
			b.append(level).append(s.charAt(i));
		return b.toString();
	}
	
	private String expandName(String s, int level) {
		if (level > 0)
			return s + "#" + level;
		return s;
	}
	
	private boolean contains(Automaton a, String par) {
		String s = a.getSingleton();
		if (s != null) 
			return contains(s, par);
		else 
			for (State p : a.getStates())
				for (Transition t : p.getTransitions()) {
					for (int i = 0; i < par.length(); i++) {
						char c = par.charAt(i);
						if (t.getMin() <= c && c <= t.getMax())
							return true;
					}
				}
		return false;
	}
	
	private boolean contains(String s, String par) {
		for (int i = 0; i < s.length(); i++)
			if (par.indexOf(s.charAt(i)) >= 0)
				return true;
		return false;
	}
}
