package dk.brics.grammar.parser;

import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import dk.brics.automaton.Automaton;
import dk.brics.grammar.EOFTerminalEntity;
import dk.brics.grammar.Entity;
import dk.brics.grammar.Grammar;
import dk.brics.grammar.GrammarException;
import dk.brics.grammar.NonterminalEntity;
import dk.brics.grammar.Production;
import dk.brics.grammar.ProductionID;
import dk.brics.grammar.RegexpTerminalEntity;
import dk.brics.grammar.StringTerminalEntity;
import dk.brics.grammar.ast.AST;
import dk.brics.grammar.ast.BranchNode;
import dk.brics.grammar.operations.GrammarChecker;
import dk.brics.misc.Chars;

/**
 * Converter from {@link AST}s representing grammars to {@link Grammar} objects.
 */
public class AST2Grammar {

	private Set<String> unused_regexps;
	
	private List<DeriveCheck> derive_checks;
	
	private static class DeriveCheck {

		String nt;
		
		String example;
		
		DeriveCheck(String nt, String example) {
			this.nt = nt;
			this.example = example;
		}
	}
	
	/** Current AST. */
	private AST ast;
	
	/**
	 * Constructs a new converter.
	 */
	public AST2Grammar() {}

	/**
	 * Converts the given AST.
	 * @param a AST
	 * @param out grammar warnings are printed to this stream if non-null 
	 *            (currently checking for unused regular expressions and issues described in {@link GrammarChecker#check(Grammar,PrintWriter)})
	 * @return grammar
	 * @throws GrammarException if the grammar is not well-defined
	 */
	public Grammar convert(AST a, PrintWriter out) throws GrammarException {
		ast = a;
		unused_regexps = new HashSet<String>();
		derive_checks = new ArrayList<DeriveCheck>();
		BranchNode n = a.getRoot();
		Map<String,Automaton> automata = new HashMap<String,Automaton>();
		Set<String> max = new HashSet<String>();
		buildRegexps(n, automata, max);
		List<Production> productions = buildProductions(n, automata, max);
		if (productions.isEmpty())
			throw new GrammarException("no productions!");
		String start = n.getBranchChild("productions").getBranchChild("production").getLeafString("nonterminal", ast);
		Grammar g = new Grammar(start, productions);
		if (out != null) {
			for (String s : unused_regexps) 
				out.println("*** unused regular expression '" + s + "'");
			new GrammarChecker().check(g, out);
		}
		if (!derive_checks.isEmpty()) {
			Parser parser = new Parser(g, null);
			for (DeriveCheck dc : derive_checks) 
				try {
					g.setStart(dc.nt);
					if (!parser.check(dc.example, null))
						throw new GrammarException("example string '" + Chars.escape(dc.example) + "' does not match nonterminal " + dc.nt);
				} finally {
					g.setStart(start);
				}
		}
		ast = null;
		unused_regexps = null;
		derive_checks = null;
		return g;
	}
	
	private void buildRegexps(BranchNode n, Map<String,Automaton> automata, Set<String>max) {
		BranchNode rs = n.getBranchChild("regexps");
		while (rs.getLabel().equals("nonempty")) {
			BranchNode r = rs.getBranchChild("regexp");
			String re = r.getLeafString("name", ast);
			if (re.equals("EOF"))
				throw new GrammarException("regular expression name 'EOF' is predefined!");
			if (automata.containsKey(re))
				throw new GrammarException("multiple definitions of regular expression '" + re + "'!");
			Automaton a = buildAutomaton(r.getBranchChild("exp"), automata);
			assert(a != null);
			automata.put(re, a);
			if (r.getBranchChild("max").getLabel().equals("present"))
				max.add(re);
			unused_regexps.add(re);
			rs = rs.getBranchChild("more");
		}
	}

	private List<Production> buildProductions(BranchNode n, Map<String,Automaton> automata, Set<String> max) {
		List<Production> productions = new ArrayList<Production>();
		BranchNode ps = n.getBranchChild("productions");
		Map<String,Integer> priority = new HashMap<String,Integer>();
		while (ps.getLabel().equals("nonempty")) {
			BranchNode p = ps.getBranchChild("production");
			String nonterminal = p.getLeafString("nonterminal", ast);
			do {
				List<Entity> entities = buildEntities(p.getBranchChild("entities"), automata, max);
				boolean unordered = p.getBranchChild("unordered").getLabel().equals("present");
				String title = null;
				BranchNode ot = p.getBranchChild("label");
				if (ot.getLabel().equals("present"))
					title = ot.getLeafString("label", ast);
				Integer pri = priority.get(nonterminal);
				if (pri == null)
					pri = 0;
				if (p.getBranchChild("priority").getLabel().equals("higher")) {
					pri--;
					priority.put(nonterminal, pri);
				}
				productions.add(new Production(nonterminal, entities, unordered, new ProductionID(title), pri));
				p = p.getBranchChild("more");
			} while (p.getLabel().equals("nonempty"));
			ps = ps.getBranchChild("more");
		}
		return productions;
	}

	private List<Entity> buildEntities(BranchNode es, Map<String,Automaton> automata, Set<String> max) {
		List<Entity> entities = new ArrayList<Entity>();
		while (es.getLabel().equals("nonempty")) {
			BranchNode e = es.getBranchChild("entity");
			Entity n = null;
			String kind = e.getLabel();
			if (kind.equals("nonterminal")) {
				String nt = e.getLeafString("nonterminal", ast);
				BranchNode ne = e.getBranchChild("labelexample");
				String nl = ne.getLabel();
				if (nl.equals("label")) 
					n = new NonterminalEntity(nt, ne.getLeafString("label", ast), null);
				else if (nl.equals("example")) {
					String example = unescape(ne.getLeafString("example", ast));
					n = new NonterminalEntity(nt, null, example);
					derive_checks.add(new DeriveCheck(nt, example));
				} else if (nl.equals("absent"))
					n = new NonterminalEntity(nt, null, null);
			} else if (kind.equals("regexp_terminal")) {
				String re = e.getLeafString("regexp", ast);
				unused_regexps.remove(re);
				BranchNode ne = e.getBranchChild("labelexample");
				String nl = ne.getLabel();
				if (re.equals("EOF")) {
					n = new EOFTerminalEntity();
					if (nl.equals("example"))
						throw new GrammarException("can't set example string for EOF terminal!");
					if (nl.equals("label"))
						throw new GrammarException("can't set label for EOF terminal!");
				} else {
					Automaton a = automata.get(re);
					if (a == null)
						throw new GrammarException("regular expression '" + re + "' not defined!");
					boolean m = max.contains(re);
					if (nl.equals("label")) 
						n = new RegexpTerminalEntity(a, m, re, ne.getLeafString("label", ast), null);
					else if (nl.equals("example")) { 
						String example = unescape(ne.getLeafString("example", ast));
						n = new RegexpTerminalEntity(a, m, re, null, example);
						if (example != null && !a.run(example))
							throw new GrammarException("example string '" + Chars.escape(example) + "' does not match regexp " + re);
					} else if (nl.equals("absent"))
						n = new RegexpTerminalEntity(a, m, re, null, null);
				}
			} else if (kind.equals("string_terminal")) 
				n = new StringTerminalEntity(unescape(e.getLeafString("string", ast)));
			entities.add(n);
			es = es.getBranchChild("more");
		}		
		return entities;
	}

	private Automaton buildAutomaton(BranchNode n, Map<String, Automaton> automata) {
		return buildAutomatonUnionExp(n.getBranchChild("e"), automata);
	}

	private Automaton buildAutomatonUnionExp(BranchNode n, Map<String, Automaton> automata) {
		Automaton a = null;
		String kind = n.getLabel();
		if (kind.equals("union")) {
			a = buildAutomatonInterExp(n.getBranchChild("e"), automata)
			    .union(buildAutomatonUnionExp(n.getBranchChild("more"), automata));
			a.minimize();
		} else if (kind.equals("other"))
			a = buildAutomatonInterExp(n.getBranchChild("e"), automata);
		assert(a != null);
		return a;
	}

	private Automaton buildAutomatonInterExp(BranchNode n, Map<String, Automaton> automata) {
		Automaton a = null;
		String kind = n.getLabel();
		if (kind.equals("inter")) {
			a = buildAutomatonConcatExp(n.getBranchChild("e"), automata)
		        .intersection(buildAutomatonInterExp(n.getBranchChild("more"), automata));			
			a.minimize();
		} else if (kind.equals("other"))
			a = buildAutomatonConcatExp(n.getBranchChild("e"), automata);
		assert(a != null);
		return a;
	}

	private Automaton buildAutomatonConcatExp(BranchNode n, Map<String, Automaton> automata) {
		Automaton a = null;
		String kind = n.getLabel();
		if (kind.equals("concat")) {
			a = buildAutomatonRepeatExp(n.getBranchChild("e"), automata)
		        .concatenate(buildAutomatonConcatExp(n.getBranchChild("more"), automata));
			a.minimize();
		} else if (kind.equals("other"))
			a = buildAutomatonRepeatExp(n.getBranchChild("e"), automata);
		assert(a != null);
		return a;
	}

	private Automaton buildAutomatonRepeatExp(BranchNode n, Map<String, Automaton> automata) {
		Automaton a = null;
		String kind = n.getLabel();
		if (kind.equals("optional"))
			a = buildAutomatonRepeatExp(n.getBranchChild("e"), automata).optional();
		else if (kind.equals("star"))
			a = buildAutomatonRepeatExp(n.getBranchChild("e"), automata).repeat();
		else if (kind.equals("plus"))
			a = buildAutomatonRepeatExp(n.getBranchChild("e"), automata).repeat(1);
		else if (kind.equals("number")) {
			int nn = Integer.parseInt(n.getLeafString("n", ast));
			a = buildAutomatonRepeatExp(n.getBranchChild("e"), automata).repeat(nn, nn);
		} else if (kind.equals("min")) {
			int nn = Integer.parseInt(n.getLeafString("n", ast));
			a = buildAutomatonRepeatExp(n.getBranchChild("e"), automata).repeat(nn);
		} else if (kind.equals("interval")) {
			int nn = Integer.parseInt(n.getLeafString("n", ast));
			int mm = Integer.parseInt(n.getLeafString("m", ast));
			a = buildAutomatonRepeatExp(n.getBranchChild("e"), automata).repeat(nn, mm);
		} else if (kind.equals("other"))
			a = buildAutomatonComplExp(n.getBranchChild("e"), automata);
		assert(a != null);
		a.minimize();
		return a;
	}

	private Automaton buildAutomatonComplExp(BranchNode n, Map<String, Automaton> automata) {
		Automaton a = null;
		String kind = n.getLabel();
		if (kind.equals("complement"))
			a = buildAutomatonComplExp(n.getBranchChild("e"), automata).complement();
		else if (kind.equals("other"))
			a = buildAutomatonCharclassExp(n.getBranchChild("e"), automata);
		assert(a != null);
		return a;
	}

	private Automaton buildAutomatonCharclassExp(BranchNode n, Map<String, Automaton> automata) {
		Automaton a = null;
		String kind = n.getLabel();
		if (kind.equals("charclass"))
			a = buildAutomatonCharclasses(n.getBranchChild("c"));
		else if (kind.equals("negativeclass"))
			a = buildAutomatonCharclasses(n.getBranchChild("c")).complement().intersection(Automaton.makeAnyChar());
		else if (kind.equals("other"))
			a = buildAutomatonSimpleExp(n.getBranchChild("e"), automata);
		assert(a != null);
		return a;
	}

	private Automaton buildAutomatonCharclasses(BranchNode n) {
		Automaton a = null;
		String kind = n.getLabel();
		if (kind.equals("first"))
			a = buildAutomatonCharclass(n.getBranchChild("c"))
			    .union(buildAutomatonCharclasses(n.getBranchChild("more")));
		else if (kind.equals("last"))
			a = buildAutomatonCharclass(n.getBranchChild("c"));
		assert(a != null);
		return a;
	}

	private Automaton buildAutomatonCharclass(BranchNode n) {
		Automaton a = null;
		String kind = n.getLabel();
		if (kind.equals("interval"))
			a = Automaton.makeCharRange(getChar(n.getBranchChild("c1")), getChar(n.getBranchChild("c2")));
		else if (kind.equals("single"))
			a = Automaton.makeChar(getChar(n.getBranchChild("c")));
		assert(a != null);
		return a;
	}
	
	private Automaton buildAutomatonSimpleExp(BranchNode n, Map<String, Automaton> automata) {
		Automaton a = null;
		String kind = n.getLabel();
		if (kind.equals("char") || kind.equals("escape"))
			a = Automaton.makeChar(getChar(n));
		else if (kind.equals("dot"))
			a = Automaton.makeAnyChar();
		else if (kind.equals("empty"))
			a = Automaton.makeEmpty();
		else if (kind.equals("all"))
			a = Automaton.makeAnyString();
		else if (kind.equals("string"))
			a = Automaton.makeString(unescape(n.getLeafString("string", ast)));
		else if (kind.equals("epsilon"))
			a = Automaton.makeEmptyString();
		else if (kind.equals("exp"))
			a = buildAutomatonUnionExp(n.getBranchChild("e"), automata);
		else if (kind.equals("named")) {
			String re = n.getLeafString("id", ast);
			if (re.equals("EOF"))
				throw new GrammarException("EOF not allowed in regular expressions!");
			a = automata.get(re);
			if (a == null)
				throw new GrammarException("regular expression '" + re + "' not defined!");
			unused_regexps.remove(re);
		} else if (kind.equals("numeric")) {
			int nn = Integer.parseInt(n.getLeafString("n", ast));
			int mm = Integer.parseInt(n.getLeafString("m", ast));
			a = Automaton.makeInterval(nn, mm, 0);
		}
		assert(a != null);
		return a;		
	}

	private String unescape(String str) {
		StringBuilder b = new StringBuilder();
		for (int i = 0; i < str.length(); i++) {
			char c = str.charAt(i);
			if (c == '\\') {
				if (str.charAt(i+1) == 'u') {
					b.append(unescapeChar(str.substring(i, i+6)));
					i += 5;
				} else {
					b.append(unescapeChar(str.substring(i, i+2)));
					i += 1;
				}
			} else
				b.append(c);
		}
		return b.toString();
	}

	private char getChar(BranchNode n) {
		Character c = null;
		String kind = n.getLabel();
		if (kind.equals("char"))
			c = n.getLeafString("c", ast).charAt(0);
		else if (kind.equals("escape")) 
			c = unescapeChar(n.getLeafString("c", ast));
		assert(c != null);
		return c;
	}

	private Character unescapeChar(String s) {
		Character c;
		char c2 = s.charAt(1);
		switch (c2) {
		case 'b':
			c = '\b';
			break;
		case 't':
			c = '\t';
			break;
		case 'n':
			c = '\n';
			break;
		case 'r':
			c = '\r';
			break;
		case 'f':
			c = '\f';
			break;
		case 'u':
			c = (char)Integer.parseInt(s.substring(2), 16);
			break;
		default:
			c = c2;
		}
		return c;
	}
}
