package dk.brics.grammar.parser;

import java.util.List;

import dk.brics.grammar.Entity;
import dk.brics.grammar.Production;
import dk.brics.grammar.ast.BranchNode;

/**
 * Parse state.
 */
class PState implements Comparable<PState> {
	
	/** Grammar production. */
	Production prod;
	
	/** Remaining entities (for unordered productions only). */
	List<Entity> remaining;
	
	/** Index of next remaining entity (for ordered productions only). */
	int next_entity;
	
	/** Index into parsed string where (partial) match starts. */
	int from;
	
	/** Index into parsed string where (partial) match ends, excluding the end point. */
	int current;
	
	/** AST for the (partial) match. */
	BranchNode ast;
	
	/** Serial number. (Used for stable comparison.) */
	int serial;
	
	/** Production of serial numbers. */
	static int counter;
	
	/**
	 * Makes new parse state.
	 */
	private PState(Production prod, List<Entity> remaining, int next_entity, int from, int current, BranchNode ast) {
		this.prod = prod;
		this.remaining = remaining;
		this.next_entity = next_entity;
		this.from = from;
		this.current = current;
		this.ast = ast;
		serial = counter++;
	}
	
	/** Makes new parse state with all entities remaining. */
	static PState makeNew(Production prod, int from, BranchNode ast) {
		return new PState(prod, prod.isUnordered() ? prod.getEntities() : null, 0, from, from, ast);
	}

	/** Makes new parse state for unordered production. */
	static PState makeNextUnordered(Production prod, List<Entity> remaining, int from, int current, BranchNode ast) {
		return new PState(prod, remaining, 0, from, current, ast);
	}
	
	/** Makes new parse state for ordered production. */
	static PState makeNextOrdered(Production prod, int next_entity, int from, int current, BranchNode ast) {
		return new PState(prod, null, next_entity, from, current, ast);
	}
	
	boolean isDone() {
		if (prod.isUnordered())
			return remaining.isEmpty();
		else
			return next_entity == prod.getEntities().size();
	}
	
	Entity getNextRemaining() { // call for ordered only
		return prod.getEntities().get(next_entity);
	}
	
	List<Entity> getRemaining() {
		if (prod.isUnordered())
			return remaining;
		else 
			return prod.getEntities().subList(next_entity, prod.getEntities().size());
	}
	
	/** Returns index into parsed string where (partial) match starts. */
	public int getFrom() {
		return from;
	}
	
	/** Returns index into parsed string where (partial) match ends, excluding the end point. */
	public int getCurrent() {
		return current;
	}
	
	@Override
	public boolean equals(Object obj) {
		if (obj instanceof PState) {
			PState s = (PState)obj;
			if (prod.isUnordered())
				return s.prod==prod && s.from == from && s.current == current 
					&& s.remaining.equals(remaining);
//					&& new HashSet<Entity>(s.remaining).equals(new HashSet<Entity>(remaining));
			else
				return s.prod==prod && s.next_entity==next_entity && s.from == from && s.current == current;
		} else
			return false;
	}
	
	@Override
	public int hashCode() {
		int r;
		if (prod.isUnordered())
			r = remaining.size();
		else
			r = next_entity;
		return prod.hashCode() + r*5 + from*3 + current*2;
	}

	public int compareTo(PState o) {
		if (prod.getPriority() > o.prod.getPriority())
			return -1;
		if (prod.getPriority() < o.prod.getPriority())
			return 1;
		if (current > o.current)
			return -1;
		if (current < o.current)
			return 1;
		return serial - o.serial; // ensures consistence with equals
	}
}
