package dk.brics.grammar.parser;

import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;

import dk.brics.automaton.Automaton;
import dk.brics.automaton.State;
import dk.brics.grammar.EOFTerminalEntity;
import dk.brics.grammar.Entity;
import dk.brics.grammar.EntityVisitor;
import dk.brics.grammar.Grammar;
import dk.brics.grammar.NonterminalEntity;
import dk.brics.grammar.Production;
import dk.brics.grammar.RegexpTerminalEntity;
import dk.brics.grammar.StringTerminalEntity;
import dk.brics.grammar.VoidEntityVisitor;
import dk.brics.grammar.ast.AST;
import dk.brics.grammar.ast.Node;
import dk.brics.grammar.ast.BranchNode;
import dk.brics.grammar.ast.LeafNode;
import dk.brics.grammar.operations.CharSet;
import dk.brics.grammar.operations.SelectFollowsFinder;
import dk.brics.grammar.operations.Unparser;
import dk.brics.misc.Properties;

/**
 * Parser.
 * <p>
 * Based on a variant of Earley's algorithm.
 */
public class Parser {
	
	/** Grammar for this parser instance. */
	private Grammar g;
	
	/** If set to true, debug info is generated. (Enable dynamically with dk.brics.grammar.parser.debug)*/
	private boolean debug = false;
	
	/** Print stream for messages, null if none. */
	private PrintWriter out;
	
	/** Flag for enabling/disabling use of SELECT/FOLLOW sets (for performance). */
	private boolean enable_selectfollows = true;
	
	/** Maximal index (into text being parsed) of successful parsing. */
	private int max_i;

	/** Pending parse states. */
	private SortedSet<PState> pending;
	
	/** Visited parse states. */
	private Set<PState> visited;

	/** Map from 'from' index and nonterminal name to list of parse states with those values. */
	private Map<Integer,Map<String,List<PState>>> ntmap;
	
	/** Map from 'from' index and nonterminal name to list of completed parse states with those values (for pending completions). */
	private Map<Integer,Map<String,List<PState>>> complmap;
	
	/** Map from nonterminal name to SELECT set. */
	private Map<String,CharSet> select;
	
	/** Map from nonterminal name to FOLLOWS set. */
	private Map<String,CharSet> follows;
	
	/** String being parsed. */
	private String x;
	
	/** Maximal memory usage during last parsing. */
	private long max_memory;
	
	/** Maximal number of parse states during last parsing. */
	private long max_states;
	
	/** Total number of parse states during last parsing. */
	private long total_states;
	
	/**
	 * Unparser.
	 */
	private Unparser unparser;
	
	/**
	 * Constructs a new parser.
	 * @param g grammar
	 * @param out print writer for verbose messages (null if none)
	 */
	public Parser(Grammar g, PrintWriter out) {
		this.g = g;
		this.out = out;
		unparser = new Unparser(g);
		checkDebugMode();
		if (enable_selectfollows) {
			select = new HashMap<String,CharSet>();
			follows = new HashMap<String,CharSet>();
			SelectFollowsFinder ff = new SelectFollowsFinder(g);
			select = ff.getNonterminalSelect();
			follows = ff.getNonterminalFollows();
			if (debug) {
				for (String n : g.getNonterminals()) {
					out.println("select(" + n + ")=" + select.get(n));
					out.println("follows(" + n + ")=" + follows.get(n));
				}
			}
		}
	}

	private void checkDebugMode() {
		debug = out != null && Properties.get("dk.brics.grammar.parser.debug");
	}
	
	/**
	 * Checks whether the given string can be parsed without syntax errors.
	 * @param x string to parse
	 * @param start start production (if null, all productions of the start nonterminal are used)
	 * @return true iff the string can be parsed by the grammar
	 */
	public boolean check(String x, Production start) {
		try {
			parse(x, null, start);
			return true;
		} catch (ParseException e) {
			return false;
		}
	}
	
	/**
	 * Parses the given string.
	 * Same as {@link #parse(String, String, Production) parse(x, file, null)}.
	 * @param x string to parse
	 * @param file path or URL of string to parse (null if unknown)
	 * @return an abstract syntax tree for x
	 * @throws ParseException in case of parse error
	 */
	public AST parse(String x, String file) throws ParseException {
		return parse(x, file, null);
	}
	
	/**
	 * Parses the given string.
	 * Same as {@link #parse(String, String, Production) parse(x, null, null)}.
	 * @param x string to parse
	 * @return an abstract syntax tree for x
	 * @throws ParseException in case of parse error
	 */
	public AST parse(String x) throws ParseException {
		return parse(x, null, null);
	}
	
	/**
	 * Parses the given string.
	 * @param x string to parse
	 * @param file path or URL of string to parse (null if unknown)
	 * @param start start production (if null, all productions of the start nonterminal are used)
	 * @return an abstract syntax tree for x
	 * @throws ParseException in case of parse error
	 */
	public AST parse(String x, String file, Production start) throws ParseException {
		checkDebugMode();
		AST res = null;
		boolean ff = enable_selectfollows;
		try {
			pending = new TreeSet<PState>();
			visited = new HashSet<PState>();
			ntmap = new HashMap<Integer,Map<String,List<PState>>>();
			complmap = new HashMap<Integer,Map<String,List<PState>>>();
			max_states = total_states = 0;
			this.x = x;
			max_i = 0;
			if (start != null) {
				enable_selectfollows = false; // must temporarily disable(/recompute) select-follows if the grammar changes
				addState(PState.makeNew(start, 0, new BranchNode(start.getID(), start.getNonterminal(), 0, 0)));
				
			} else
				for (Production p : g.getProductions(g.getStart())) 
					addState(PState.makeNew(p, 0, new BranchNode(p.getID(), p.getNonterminal(), 0, 0)));
			while (!pending.isEmpty()) {
				PState s = pending.first();
				pending.remove(s);
				if (debug)
					out.println("processing " + s.prod.print(s.getRemaining()) + " from index " + s.from + " to " + s.current);
				if (s.isDone()) { // end of production
					if (debug) 
						out.println("reducing " + s.prod.print(s.getRemaining()) + " from index " + s.from + " to " + s.current);
					if ((start == s.prod || (start == null && s.prod.getNonterminal().equals(g.getStart()))) 
							&& s.from == 0 && s.current == x.length()) {
						res = new AST(s.ast, x);
						break;
					}
					addCompleted(s);
					Map<String,List<PState>> cm = ntmap.get(s.from);
					if (cm != null) {
						List<PState> cl = cm.get(s.prod.getNonterminal());
						if (cl != null)
							for (PState r : new ArrayList<PState>(cl))
								if (r.prod.isUnordered()) {
									for (Entity e : r.getRemaining())
										if (e instanceof NonterminalEntity && ((NonterminalEntity)e).getNonterminal().equals(s.prod.getNonterminal()))
											complete(s, r, (NonterminalEntity)e);
								} else
									complete(s, r, (NonterminalEntity)r.getNextRemaining());
					}
				} else 
					if (s.prod.isUnordered())
						for (Entity e : s.getRemaining())
							parseEntity(s, e);
					else
						parseEntity(s, s.getNextRemaining());
			}
			if (res == null)
				throw new ParseException(file, x, max_i);
		} finally {
			pending = null;
			visited = null;
			ntmap = null;
			complmap = null;
			this.x = null;
			enable_selectfollows = ff;
			if (debug) {
				out.println("max_states = " + max_states);
				out.println("total_states = " + total_states);
			}
		}
		return res;
	}

	/** 
	 * Reached end of production. Remember that for later in case we need to complete more.
	 * @param s parse state that has reached end of production  
	 */
	private void addCompleted(PState s) {
		Map<String,List<PState>> mm = complmap.get(s.from);
		if (mm == null) {
			mm = new HashMap<String,List<PState>>();
			complmap.put(s.from, mm);
		}
		List<PState> ml = mm.get(s.prod.getNonterminal());
		if (ml == null) {
			ml = new ArrayList<PState>();
			mm.put(s.prod.getNonterminal(), ml);
		}
		ml.add(s);
	}
	
	/**
	 * Completes by processing states where the next entity is the relevant nonterminal.
	 * @param s parse state that has reached end of production of nonterminal <code>e</code>
	 * @param r parse state whose next entity is the nonterminal <code>e</code>
	 * @param e the next entity
	 * @throws ParseException in case of parse error
	 */
	private void complete(PState s, PState r, NonterminalEntity e) throws ParseException {
		BranchNode ast = null;
		boolean ok = true;
		if (r.prod.isUnordered() && s.getFrom() == s.getCurrent()) { // avoid combinatorial blowup by nullable entities in unordered productions
			Entity next = r.getRemaining().get(0);
			if (!(next instanceof NonterminalEntity) || !((NonterminalEntity)next).getNonterminal().equals(e.getNonterminal())) {
				if (debug)
					out.println("cancelling epsilon completion of " + e.getNonterminal() + " for " + r.prod.print(r.getRemaining()));
				ok = false;
			}
		}
		if (ok) 
			if (e.isExplicitlyLabeled()) {
				ok = checkEquality(r.ast, e.getLabel(), s.ast);
				ast = new BranchNode(r.ast, e.getLabel(), s.ast);
			} else
				ast = new BranchNode(r.ast, s.current);
		if (ok) {
			PState n;
			if (r.prod.isUnordered()) {
				List<Entity> newremaining = new ArrayList<Entity>(r.getRemaining());
				newremaining.remove(e);
				n = PState.makeNextUnordered(r.prod, newremaining, r.from, s.current, ast);
			} else
				n = PState.makeNextOrdered(r.prod, r.next_entity + 1, r.from, s.current, ast);
			addState(n);
		}
	}
	
	/**
	 * Checks that the substring represented by <code>edge</code> of the AST node <code>r</code> 
	 * is equal to the substring represented by the AST node <code>n</code>.
	 * @param r an existing AST node
	 * @param edge an outgoing edge of <code>r</code>, pointing to some AST node
	 * @param n a potential new AST node
	 * @return true if the substrings are equal
	 */
	private boolean checkEquality(BranchNode r, String edge, Node n) {
		boolean ok = true;
		Node m = r.getChild(edge);
		if (m != null) {
			String ns = unparser.unparse(n, x);
			String ms = unparser.unparse(m, x);
			if (!ns.equals(ms))
				ok = false;
		}
		return ok;
	}
	
	/**
	 * Processes a parse state that has not reached the end of the production.
	 * @param s parse state to process
	 * @param e next entity of <code>s</code>
	 * @throws ParseException in case of parse error
	 */
	private void parseEntity(final PState s, final Entity e) throws ParseException {
		ArrayList<Entity> nr = null;
		if (s.prod.isUnordered()) {
			nr = new ArrayList<Entity>(s.getRemaining());
			nr.remove(e);
		}
		final ArrayList<Entity> newremaining = nr;
		final List<PState> newstates = new ArrayList<PState>();
		e.visitBy(new EntityVisitor<Object>() {
			public Object visitStringTerminalEntity(StringTerminalEntity t) {
				String str = t.getString();
				int len = str.length();
				if (s.current + len <= x.length() && x.substring(s.current, s.current + len).equals(str)) {
					if (debug)
						out.println("matched string terminal " + str);
					BranchNode ast;
					boolean ok = true;
					if (e.isExplicitlyLabeled()) {
						Node m = new LeafNode(s.current, s.current + len);
						ok = checkEquality(s.ast, e.getLabel(), m);
						ast = new BranchNode(s.ast, e.getLabel(), m);
					} else
						ast = new BranchNode(s.ast, s.current + len);
					if (ok) {
						PState n;
						if (s.prod.isUnordered())
							n = PState.makeNextUnordered(s.prod, newremaining, s.from, s.current + len, ast);
						else
							n = PState.makeNextOrdered(s.prod, s.next_entity + 1, s.from, s.current + len, ast);
						newstates.add(n);
					}
				}
				return null;
			}
			public Object visitRegexpTerminalEntity(RegexpTerminalEntity t) {
				List<PState> news = new ArrayList<PState>();
				Automaton a = t.getAutomaton();
				State q = a.getInitialState();
				int len = 0;
				while (q != null) {
					if (q.isAccept()) {
						if (debug)
							out.println("matched regexp terminal " + t.getAutomatonName());
						BranchNode ast;
						boolean ok = true;
						if (e.isExplicitlyLabeled()) {
							Node m = new LeafNode(s.current, s.current + len);
							ok = checkEquality(s.ast, e.getLabel(), m);
							ast = new BranchNode(s.ast, e.getLabel(), m);
						} else 
							ast = new BranchNode(s.ast, s.current + len);
						if (ok) {
							PState n;
							if (s.prod.isUnordered())
								n = PState.makeNextUnordered(s.prod, newremaining, s.from, s.current + len, ast);
							else
								n = PState.makeNextOrdered(s.prod, s.next_entity + 1, s.from, s.current + len, ast);
							news.add(n);
						}
					}
					if (s.current + len >= x.length()) 
						break;
					q = q.step(x.charAt(s.current + len++));
				}
				if (t.isMax() && !news.isEmpty())
					newstates.add(news.get(news.size() - 1));
				else
					for (PState ns: news)
						newstates.add(ns);
				return null;
			}
			public Object visitEOFTerminalEntity(EOFTerminalEntity t) {
				if (s.current == x.length()) {
					if (debug)
						out.println("matched EOF");
					PState n;
					if (s.prod.isUnordered())
						n = PState.makeNextUnordered(s.prod, newremaining, s.from, s.current, s.ast);
					else
						n = PState.makeNextOrdered(s.prod, s.next_entity + 1, s.from, s.current, s.ast);
					newstates.add(n);
				}
				return null;
			}
			public Object visitNonterminalEntity(NonterminalEntity t) {
				Collection<Production> pp = g.getProductions(t.getNonterminal());
				if (pp != null)
					for (Production q : pp) 
						newstates.add(PState.makeNew(q, s.current, new BranchNode(q.getID(), q.getNonterminal(), s.current, s.current)));
				return null;
			}
		});
		for (PState n : newstates)
			addState(n);
	}
	
	/**
	 * Adds new parse state.
	 * @param n parse state to add
	 * @throws ParseException in case of parse error
	 */
	private void addState(PState n) throws ParseException {
		if (n.current > max_i)
			max_i = n.current;
		if (!visited.contains(n)) {
			if (!isApplicable(n)) {
				if (debug)
					out.println("nonmatching lookahead for " + n.prod.print(n.getRemaining()) + " from index " + n.from + " to " + n.current);
			} else {
				if (debug)
					out.println("adding " + n.prod.print(n.getRemaining()) + " from index " + n.from + " to pending[" + n.current + "]");
				visited.add(n);
				total_states++;
				pending.add(n);
				if (pending.size() > max_states)
					max_states = pending.size();
				Runtime rt = Runtime.getRuntime();
				long mem = rt.totalMemory() - rt.freeMemory();
				if (mem > max_memory)
					max_memory = mem;
				handlePendingCompletes(n);
			}
		}
	}
	
	/**
	 * Tries to weed out hopeless parse states (using SELECT/FOLLOWS). 
	 * @param n parse state to check
	 * @return false if definitely hopeless, true if it might have a chance
	 */
	private boolean isApplicable(PState n) {
		if (n.isDone()) {
			if (enable_selectfollows) {
				CharSet f = follows.get(n.prod.getNonterminal());
				if (f != null && 
						((n.current < x.length() && !f.contains(x.charAt(n.current))) ||
								(n.current == x.length() && !f.containsEOF())))
					return false;
			}
			return true;
		} else {
			if (n.prod.isUnordered()) { 
				for (Entity e : n.getRemaining())
					if (isApplicableEntity(n, e))
						return true;
				return false;
			} else
				return isApplicableEntity(n, n.getNextRemaining());
		}
	}

	private Boolean isApplicableEntity(final PState n, Entity e) {
		return e.visitBy(new EntityVisitor<Boolean>() {
			public Boolean visitNonterminalEntity(NonterminalEntity e) {
				if (enable_selectfollows) {
					CharSet f = select.get(e.getNonterminal());
					if (f != null && 
							((n.current < x.length() && !f.contains(x.charAt(n.current))) ||
									(n.current == x.length() && !f.containsEOF())))
						return false;
				}
				return true;
			}
			public Boolean visitRegexpTerminalEntity(RegexpTerminalEntity e) {
				State q = e.getAutomaton().getInitialState();
				int len = 0;
				while (q != null) {
					if (q.isAccept() || len > 100) 
						break;
					if (n.current + len >= x.length())
						return false;
					q = q.step(x.charAt(n.current + len++));
				}
				return q != null;
			}
			public Boolean visitStringTerminalEntity(StringTerminalEntity e) {
				String str = e.getString();
				int len = str.length();
				return n.current + len <= x.length() && x.substring(n.current, n.current + len).equals(str); 
			}
			public Boolean visitEOFTerminalEntity(EOFTerminalEntity e) {
				return n.current == x.length();
			}
		});
	}
	
	/**
	 * Checks for pending completes. This might trigger additional parse states to be added.
	 * @param n newly added parse state
	 * @throws ParseException in case of parse error
	 */
	private void handlePendingCompletes(PState n) throws ParseException {
		if (!n.isDone()) 
			if (n.prod.isUnordered()) 
				for (Entity e : n.getRemaining())
					handlePendingCompletes(n, e);
			else
				handlePendingCompletes(n, n.getNextRemaining());
	}

	/**
	 * Checks for pending completes. This might trigger additional parse states to be added.
	 * @param n newly added parse state
	 * @param e next entity in <code>n</code>
	 * @throws ParseException in case of parse error
	 */
	private void handlePendingCompletes(final PState n, Entity e) throws ParseException {
		final List<PState> cs = new ArrayList<PState>();
		e.visitBy(new VoidEntityVisitor() {
			@Override public void visitNonterminal(NonterminalEntity e) {
				// add to clists
				Map<String,List<PState>> cm = ntmap.get(n.current);
				if (cm == null) {
					cm = new HashMap<String,List<PState>>();
					ntmap.put(n.current, cm);
				}
				List<PState> cl = cm.get(e.getNonterminal());
				if (cl == null) {
					cl = new ArrayList<PState>();
					cm.put(e.getNonterminal(), cl);
				}
				cl.add(n);
				// check pending completions
				Map<String,List<PState>> mm = complmap.get(n.current);
				if (mm != null) {
					List<PState> ml = mm.get(e.getNonterminal());
					if (ml != null)
						cs.addAll(ml);
				}
			}
		});
		for (PState s : cs)
			complete(s, n, (NonterminalEntity)e);
	}

	/**
	 * Returns maximal memory usage during last parsing.
	 * (Note: garbage collection might reduce this number.)
	 * @return memory usage
	 */
	public long getMaxMemory() {
		return max_memory;
	}

	/**
	 * Returns maximal number of parse states during last parsing.
	 * @return number of parse states
	 */
	public long getMaxStates() {
		return max_states;
	}

	/**
	 * Returns total number of parse states during last parsing.
	 * @return number of parse states
	 */
	public long getTotalStates() {
		return total_states;
	}
}
